# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os
import pkg_resources

from .psptool import PSPTool
from .utils import ObligingArgumentParser, PrintHelper
from .entry import PubkeyEntry, HeaderEntry
from .crypto import PrivateKeyDict

from argparse import RawTextHelpFormatter, SUPPRESS


def main():
    # CLI stuff to create a PSPTool object and interact with it
    parser = ObligingArgumentParser(description='Display, extract, and manipulate AMD PSP firmware inside BIOS ROMs.\n',
                                    formatter_class=RawTextHelpFormatter, add_help=False)

    parser.add_argument('file', help='Binary file to be parsed for PSP firmware', nargs='?')
    parser.add_argument('-h', '--help', action='help', help=SUPPRESS)
    parser.add_argument('-v', '--verbose', help=SUPPRESS, action='store_true')

    parser.add_argument('-r', '--rom-index', help=SUPPRESS, type=int, default=0)
    parser.add_argument('-d', '--directory-index', help=SUPPRESS, type=int)
    parser.add_argument('-e', '--entry-index', help=SUPPRESS, type=int)
    parser.add_argument('-s', '--subfile', help=SUPPRESS)
    parser.add_argument('-o', '--outfile', help=SUPPRESS)
    parser.add_argument('-u', '--decompress', help=SUPPRESS, action='store_true')
    parser.add_argument('-c', '--decrypt', help=SUPPRESS, action='store_true')
    parser.add_argument('-k', '--pem-key', help=SUPPRESS, action='store_true')
    parser.add_argument('-n', '--no-duplicates', help=SUPPRESS, action='store_true')
    parser.add_argument('-j', '--json', help=SUPPRESS, action='store_true')
    parser.add_argument('-t', '--key-tree', help=SUPPRESS, action='store_true')
    parser.add_argument('-p', '--privkeystub', help=SUPPRESS)
    parser.add_argument('-a', '--privkeypass', help=SUPPRESS)

    action = parser.add_mutually_exclusive_group(required=False)

    action.add_argument('-V', '--version', action='store_true')

    action.add_argument('-E', '--entries', help='\n'.join([
        'Default: Parse and display PSP firmware entries.',
        '[-n] [-j] [-t]',
        '',
        '-n:      list unique entries only ordered by their offset',
        '-j:      output in JSON format instead of tables',
        '-t:      print tree of all signed entities and their certifying keys',
        '', '']), action='store_true')

    action.add_argument('-X', '--extract-entry', help='\n'.join([
        'Extract one or more PSP firmware entries.',
        '[-d idx [-e idx]] [-n] [-u] [-c] [-k] [-o outfile]',
        '',
        '-r idx:  specifies rom_index (default: 0)',
        '-d idx:  specifies directory_index (default: all directories)',
        '-e idx:  specifies entry_index (default: all entries)',
        '-n:      skip duplicate entries and extract unique entries only',
        '-u:      uncompress compressed entries',
        '-c:      try to decrypt entries',
        '-k:      convert pubkeys into PEM format',
        '-o file: specifies outfile/outdir (default: stdout/{file}_extracted)',
        '', '']), action='store_true')

    action.add_argument('-R', '--replace-entry', help='\n'.join([
        'Copy a new entry (including header and signature) into the',
        'ROM file and update metadata accordingly.',
        '-d idx -e idx -s subfile -o outfile [-p file-stub] [-a pass]',
        '',
        '-r idx:  specifies rom_index (default: 0)',
        '-d idx:  specifies directory_index',
        '-e idx:  specifies entry_index',
        '-s file: specifies subfile (i.e. the new entry contents)',
        '-o file: specifies outfile',
        '-p file: specifies file-stub (e.g. \'keys/id\') for the re-signing keys',
        '-a pass: specifies password for the re-signing keys'
        '', '']), action='store_true')

    args = parser.parse_args()
    ph = PrintHelper(args.verbose)

    if args.version:
        print(pkg_resources.get_distribution("psptool").version)
        sys.exit(0)
    elif not args.file:
        parser.print_help(sys.stderr)
        sys.exit(0)

    psp = PSPTool.from_file(args.file, verbose=args.verbose)
    output = None

    if args.extract_entry:
        if args.directory_index is not None and args.entry_index is not None:
            entry = psp.blob.roms[args.rom_index].directories[args.directory_index].entries[args.entry_index]

            if args.decompress:
                if not entry.compressed:
                    ph.print_error_and_exit(f'Entry is not compressed {entry.get_readable_type()}')
                output = entry.get_signed_bytes()
            elif args.decrypt:
                if not entry.encrypted:
                    ph.print_error_and_exit(f'Entry is not encrypted {entry.get_readable_type()}')
                output = entry.to_decrypted_entry_bytes()
            elif args.pem_key:
                output = entry.get_pem_encoded()
            else:
                output = entry.get_bytes()

        else:
            if args.entry_index is None:  # if neither directory_index nor entry_index are specified
                if args.directory_index is not None:
                    directories = [psp.blob.roms[args.rom_index].directories[args.directory_index]]
                else:
                    directories = psp.blob.roms[args.rom_index].directories

                if args.no_duplicates is False:
                    outdir = args.outfile or f'./{psp.filename}_extracted'
                    for dir_index, directory in enumerate(directories):
                        for entry_index, entry in enumerate(directory.entries):
                            if args.decompress and type(entry) is HeaderEntry:
                                out_bytes = entry.get_signed_bytes()
                            elif args.decrypt and type(entry) is HeaderEntry:
                                out_bytes = entry.get_decrypted()
                            elif args.pem_key and type(entry) is PubkeyEntry:
                                out_bytes = entry.get_pem_encoded()
                            else:
                                out_bytes = entry.get_bytes()

                            outpath = outdir + '/d%.2d_e%.2d_%s' % (dir_index, entry_index, entry.get_readable_type())
                            if type(entry) is HeaderEntry:
                                outpath += f'_{entry.get_readable_version()}'

                            os.makedirs(os.path.dirname(outpath), exist_ok=True)
                            with open(outpath, 'wb') as f:
                                f.write(out_bytes)
                    ph.print_info(f"Extracted all entries to {outdir}")
                else:  # no_duplicates is True
                    for entry in psp.blob.roms[args.rom_index].unique_entries:
                        if args.decompress and type(entry) is HeaderEntry:
                            out_bytes = entry.get_signed_bytes()
                        elif args.decrypt and type(entry) is HeaderEntry:
                            out_bytes = entry.get_decrypted()
                        elif args.pem_key and type(entry) is PubkeyEntry:
                            out_bytes = entry.get_pem_encoded()
                        else:
                            out_bytes = entry.get_bytes()

                        outdir = args.outfile or f'./{psp.filename}_unique_extracted'
                        outpath = outdir + '/%s' % (entry.get_readable_type())

                        if type(entry) is HeaderEntry:
                            outpath += f'_{entry.get_readable_version()}'

                        os.makedirs(os.path.dirname(outpath), exist_ok=True)
                        with open(outpath, 'wb') as f:
                            f.write(out_bytes)
            else:
                parser.print_help(sys.stderr)

    elif args.replace_entry:
        if args.directory_index is not None and args.entry_index is not None and args.outfile is not None:
            entry = psp.blob.roms[args.rom_index].directories[args.directory_index].entries[args.entry_index]

            # Substituting an entry is actually optional to allow plain re-signs
            if args.subfile is not None:
                with open(args.subfile, 'rb') as f:
                    sub_binary = f.read()
                # Keep the existing entry's address, but adapt its size
                entry.move_buffer(entry.get_address(), len(sub_binary))
                entry.set_bytes(0, len(sub_binary), sub_binary)

            privkeys = None
            if args.privkeystub:
                privkeys = PrivateKeyDict.read_from_files(args.privkeystub, args.privkeypass)

            if hasattr(entry, 'signed_entity') and entry.signed_entity:
                entry.signed_entity.resign_and_replace(privkeys=privkeys, recursive=True)
            else:
                ph.print_warning("Did not resign anything since target entry is not signed")

            psp.to_file(args.outfile)

            if privkeys:
                privkeys.save_to_files(args.privkeystub, args.privkeypass)
        else:
            parser.print_help(sys.stderr)
    else:
        if args.json:
            psp.ls_json(verbose=args.verbose)
        elif args.key_tree:
            psp.cert_tree.print_key_tree()
        elif args.no_duplicates:
            psp.ls_entries(verbose=args.verbose)
        else:
            psp.ls(verbose=args.verbose)

    # Output handling (stdout or outfile)
    if output is not None:
        if args.outfile is None:
            sys.stdout.buffer.write(output)
        else:
            with open(args.outfile, 'wb') as f:
                f.write(output)


if __name__ == '__main__':
    main()



# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .crypto import PublicKey


class NoCertifyingKey(Exception):
    def __init__(self, signed_entity):
        self.signed_entity = signed_entity

    def __str__(self):
        key_id = self.signed_entity.certifying_key
        return f'There is no key with id {key_id.as_string()}, so {self.signed_entity} could not be verified!'


class SignatureInvalid(Exception):
    def __init__(self, signed_entity, pubkey: PublicKey):
        self.signed_entity = signed_entity
        self.pubkey = pubkey

    def __str__(self):
        return f'Signature for {self.signed_entity} is not signed by {self.pubkey}!'


class SignatureInconsistent(SignatureInvalid):
    def __str__(self):
        return f'Signature for {self.signed_entity} was verified successfully at least once, but could not be ' \
               f'verified by {self.pubkey}!'



class NonUniqueSignedEntity(Exception):
    def __init__(self, existing, new):
        self.existing = existing
        self.new = new

    def __str__(self):
        start = self.existing.body.get_address()
        end = start + self.existing.body.buffer_size
        return f'There was anlready a SignedEntity at {hex(start)}:{hex(end)} (existing={self.existing}, new={self.new})!'


class NonUniquePublicKeyEntity(Exception):
    def __init__(self, existing, new):
        self.existing = existing
        self.new = new

    def __str__(self):
        start = self.existing.key_id.get_address()
        return f'There was anlready a PublicKeyEntity with key_id at {hex(start)} (existing={self.existing}, new={self.new})!'




# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .utils import NestedBuffer
from .directory import Directory

from typing import List


ZEN_GENERATION_IDS = {'Zen 1': [b'\x00\x09\xBC', b'\x00\x0A\xBC'],
                      'Zen 2': [b'\x05\x0B\xBC', b'\x01\x0A\xBC'],
                      'Zen 3': [b'\x01\x0C\xBC', b'\x00\x0C\xBC']}


class EmptyFet(Exception):
    pass


class Fet(NestedBuffer):
    def __init__(self, parent_rom, fet_offset: int, psptool):

        self.rom = parent_rom
        self.psptool = psptool
        self.directories: List[Directory] = []

        super().__init__(
            self.rom,
            self._determine_size(fet_offset),
            buffer_offset=fet_offset
        )

        self._parse_entry_table()

    def __repr__(self):
        return f'Fet(len(directories)={len(self.directories)})'

    def _determine_size(self, fet_offset):
        size = 0
        step_size = 4
        end_sequence = 4 * b'\xff\xff\xff\xff'

        while self.rom[
              (fet_offset + size)
              :(fet_offset + size + len(end_sequence))
              ] != end_sequence:
            size += step_size

        if size <= 0:
            raise EmptyFet()
        return size

    def _create_directory(self, addr, magic, zen_generation):
        # todo: move this responsibility to the parent ROM
        if magic == b'$PSP':
            type_ = "PSP"
        elif magic == b'$BHD':
            type_ = "BIOS"
        elif magic == b'\xff\xff\xff\xff':
            self.psptool.ph.print_warning(f"Empty FET entry at ROM address {hex(addr)}")
            return
        else:
            self.psptool.ph.print_warning(f"Unknown FET entry with magic {magic} at ROM address {hex(addr)}")
            return
        dir_ = Directory(self.rom, addr, type_, self.psptool, zen_generation)
        self.directories.append(dir_)
        if dir_.secondary_directory_address is not None:
            self.directories.append(
                Directory(self.rom, dir_.secondary_directory_address, 'secondary', self.psptool, zen_generation)
            )

    def _parse_entry_table(self):
        entries = self.get_chunks(4, 4)
        for _index, entry in enumerate(entries):
            rom_addr = int.from_bytes(entry, 'little')
            # TODO: Why is 0xFFFFFFFe a possible value here?
            if rom_addr in [0x0, 0xFFFFFFFF, 0xFFFFFFFe]:
                continue
            rom_addr &= self.rom.addr_mask
            try:
                dir_magic = self.rom[rom_addr:rom_addr + 4]
            except:
                self.psptool.ph.print_warning(f"FET entry 0x{rom_addr:x} not found or invalid, skipping ...")
                continue
            if dir_magic == b'2PSP' or dir_magic == b'2BHD':
                combo_results = self._parse_combo_dir(rom_addr)
                for result in combo_results:
                    combo_address = result[0]
                    combo_zen_gen = result[1]
                    dir_magic = self.rom[combo_address:combo_address + 4]
                    self._create_directory(combo_address, dir_magic, combo_zen_gen)
            elif dir_magic == b'$PSP':
                self._create_directory(rom_addr, dir_magic, zen_generation='unknown')
            else:
                self._create_directory(rom_addr, dir_magic, zen_generation='unknown')
                pass

    def _parse_combo_dir(self, dir_addr):
        results = []
        no_of_entries = int.from_bytes(self.rom[dir_addr + 8: dir_addr + 0xc],
                                    'little')
        combo_dir = self.rom[dir_addr: dir_addr + 16 * (no_of_entries + 2)]

        # Combo dir entries seem to begin at offset 0x20, make sure we don't
        # miss directories that don't adhere to that rule
        assert(combo_dir[0x10:0x20] == (b'\x00' * 16))

        for i in range(2, no_of_entries+2):
            entry = combo_dir[i * 16 + 0x8: i * 16 + 0xc]
            entry_addr = int.from_bytes(entry, 'little')
            if entry_addr in [0, 0xFFFFFFFF]:
                continue
            entry_addr &= self.rom.addr_mask
            # entry_addr += self.blob_offset
            zen_generation_id = combo_dir[i*16+5:i*16+8]
            zen_generation = 'unknown'
            for possible_zen_generation in ZEN_GENERATION_IDS:
                if zen_generation_id in ZEN_GENERATION_IDS[possible_zen_generation]:
                    zen_generation = possible_zen_generation

            results.append((entry_addr, zen_generation))

        return results



# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct

from .utils import NestedBuffer, chunker, fletcher32
from .entry import Entry, BIOS_ENTRY_TYPES

from typing import List


class Directory(NestedBuffer):
    ENTRY_FIELDS = [
        'type',
        'size',
        'offset',
        'rsv0',
        'rsv1',
        'rsv2'
    ]

    _HEADER_SIZES = {
        b'$PSP': 4 * 4,
        b'$PL2': 4 * 4,
        b'$BHD': 4 * 4,
        b'$BL2': 4 * 4
    }

    _ENTRY_SIZES = {
        b'$PSP': 4 * 4,
        b'$PL2': 4 * 4,
        b'$BHD': 4 * 6,
        b'$BL2': 4 * 6
    }

    _ENTRY_TYPES_SECONDARY_DIR = [0x40, 0x70]
    _ENTRY_TYPES_PUBKEY = [0x0, 0x9, 0xa, 0x5, 0xd]

    def __init__(self, parent_rom, rom_address: int, type_: str, psptool, zen_generation='unknown'):
        self.rom = parent_rom

        # The offset of this directory as specified in the FET
        self.buffer_offset = rom_address

        self.psptool = psptool
        self.checksum = None
        self._count = None

        self.zen_generation = zen_generation

        # a directory must parse itself before it knows its size and can initialize its buffer
        self._parse_header()

        super().__init__(self.rom, self.buffer_size, buffer_offset=self.buffer_offset)

        self.type = type_
        self.entries: List[Entry] = []

        self._entry_size = self._ENTRY_SIZES[self.magic]

        self._parse_entries()

        # check entries for a link to a secondary directory (i.e. a continuation of this directory)
        self.secondary_directory_address = None
        for entry in self.entries:
            if entry.type in self._ENTRY_TYPES_SECONDARY_DIR:
                # print_warning(f"Secondary dir at 0x{entry.buffer_offset:x}")
                self.secondary_directory_address = entry.buffer_offset

        self.verify_checksum()

    def __repr__(self):
        return f'Directory(address={hex(self.get_address())}, type={self.type}, magic={self.magic}, count={self.count})'

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, value):
        self._count = value

        # update binary representation
        self.header[8:12] = struct.pack('<I', self.count)
        self.update_checksum()

    def _parse_header(self):
        # ugly to do this manually, but we do not know our size yet
        self._count = int.from_bytes(
            self.rom[self.buffer_offset + 8: self.buffer_offset + 12],
            'little'
        )
        self.magic = self.rom.get_bytes(self.buffer_offset, 4)

        self.header = NestedBuffer(
            self,
            self._HEADER_SIZES[self.magic]
        )
        self.body = NestedBuffer(
            self,
            self._ENTRY_SIZES[self.magic] * self._count,
            buffer_offset=self._HEADER_SIZES[self.magic]
        )

        self.buffer_size = len(self.header) + len(self.body)
        self.checksum = NestedBuffer(self, 4, 4)

    def _parse_entries(self):
        self.entries = []

        for entry_bytes in self.body.get_chunks(self._entry_size):
            entry_fields = {}
            for key, word in zip(self.ENTRY_FIELDS, chunker(entry_bytes, 4)):
                entry_fields[key] = struct.unpack('<I', word)[0]

            entry_fields['offset'] &= self.rom.addr_mask
            destination = None

            if entry_fields['type'] in BIOS_ENTRY_TYPES:
                destination = struct.unpack('<Q', entry_bytes[0x10:0x18])[0]

            try:
                entry = Entry.from_fields(self, self.parent_buffer,
                                          entry_fields['type'],
                                          entry_fields['size'],
                                          entry_fields['offset'],
                                          self.rom,
                                          self.psptool,
                                          destination=destination)
            except:
                self.psptool.ph.print_warning(f"Entry of {self} at {hex(entry_fields['offset'])} cannot be parsed")
                continue

            for existing_entry in self.rom.unique_entries:
                if entry == existing_entry:
                    existing_entry.references.append(self)

            self.entries.append(entry)
            self.rom.unique_entries.add(entry)

    def verify_checksum(self):
        data = self[0x8:]
        checksum = self.checksum.get_bytes()
        calculated_checksum = fletcher32(data)
        if checksum == calculated_checksum:
            return
        self.psptool.ph.print_warning(f"Could not verify checksum for directory {self}")

    def update_checksum(self):
        data = self[0x8:]  # checksum is calculated from after the checksum field in the header
        self.checksum.set_bytes(0, 4, fletcher32(data))

    def update_entry_fields(self, entry: Entry, type_, size, offset):
        entry_index = None
        for index, my_entry in enumerate(self.entries):
            if my_entry.type == entry.type:
                entry_index = index
                break

        assert(entry_index is not None)

        # apparently this masking is still needed for the PSP to parse stuff correctly
        offset |= 0xFF000000

        # update type, size, offset (but not rsv0 (nor rsv1 nor rsv2 in $BHD/$BLD directories))
        entry_bytes = b''.join([struct.pack('<I', value) for value in [type_, size, offset]])

        self.body.set_bytes(
            self._entry_size * entry_index,
            len(entry_bytes),
            entry_bytes
        )

        self.update_checksum()



# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import string
import struct

from .utils import NestedBuffer
from .utils import shannon
from .utils import zlib_decompress, zlib_compress
from .utils import decrypt
from .utils import round_to_int
from .crypto import KeyId, Signature, ReversedSignature, PrivateKey

from enum import Enum

from binascii import hexlify
from math import ceil
from hashlib import md5, sha256

BIOS_ENTRY_TYPES = [0x10062, 0x30062]


class Entry(NestedBuffer):
    ENTRY_ALIGNMENT = 0x10

    UNWRAPPED_IKEK_ZEN_PLUS = b'\x4c\x77\x63\x65\x32\xfe\x4c\x6f\xd6\xb9\xd6\xd7\xb5\x1e\xde\x59'
    HASH_IKEK_ZEN_PLUS = b'\xe2\x84\xda\xe0\x6e\x58\x01\x04\xfa\x6e\x8e\x6b\x58\x68\x8a\x0c'

    UNWRAPPED_IKEK_ZEN = b'\x49\x1e\x40\x1a\x40\x1e\xc1\xb2\x28\x46\x00\xf0\x99\xfd\xe8\x68'
    HASH_IKEK_ZEN = b'\x47\x23\xa8\x52\x03\x38\xbd\x2e\xac\x5f\xae\x9c\x2c\xb5\x92\x5b'

    DIRECTORY_ENTRY_TYPES = {
        0x00: 'AMD_PUBLIC_KEY',
        0x01: 'PSP_FW_BOOT_LOADER',
        0x02: 'PSP_FW_TRUSTED_OS',
        0x03: 'PSP_FW_RECOVERY_BOOT_LOADER',
        0x04: 'PSP_NV_DATA',
        0x05: 'BIOS_PUBLIC_KEY',
        0x06: 'BIOS_RTM_FIRMWARE',
        0x07: 'BIOS_RTM_SIGNATURE',
        0x08: 'SMU_OFFCHIP_FW',
        0x09: 'SEC_DBG_PUBLIC_KEY',
        0x0A: 'OEM_PSP_FW_PUBLIC_KEY',
        0x0B: 'SOFT_FUSE_CHAIN_01',
        0x0C: 'PSP_BOOT_TIME_TRUSTLETS',
        0x0D: 'PSP_BOOT_TIME_TRUSTLETS_KEY',
        0x10: 'PSP_AGESA_RESUME_FW',
        0x12: 'SMU_OFF_CHIP_FW_2',
        0x13: 'DEBUG_UNLOCK',
        0x1A: 'PSP_S3_NV_DATA',
        0x21: 'WRAPPED_IKEK',
        0x22: 'TOKEN_UNLOCK',
        0x24: 'SEC_GASKET',
        0x25: 'MP2_FW',
        0x26: 'MP2_FW_2',
        0x27: 'USER_MODE_UNIT_TEST',
        0x28: 'DRIVER_ENTRIES',
        0x29: 'KVM_IMAGE',
        0x2A: 'MP5_FW',
        0x2D: 'S0I3_DRIVER',
        0x30: 'ABL0',
        0x31: 'ABL1',
        0x32: 'ABL2',
        0x33: 'ABL3',
        0x34: 'ABL4',
        0x35: 'ABL5',
        0x36: 'ABL6',
        0x37: 'ABL7',
        0x38: 'SEV_DATA',
        0x39: 'SEV_CODE',
        0x3A: 'FW_PSP_WHITELIST',
        # 0x40: 'FW_L2_PTR',
        0x41: 'FW_IMC',
        0x42: 'FW_GEC',
        # 0x43: 'FW_XHCI',
        0x44: 'FW_INVALID',
        0x46: 'ANOTHER_FET',
        0x50: 'KEY_DATABASE',
        0x5f: 'FW_PSP_SMUSCS',
        0x60: 'APCB',
        0x61: 'APOB',
        0x62: 'FW_XHCI',
        0x63: 'APOB_NV_COPY',
        0x64: 'PMU_CODE',
        0x65: 'PMU_DATA',
        0x66: 'MICROCODE_PATCH',
        0x67: 'CORE_MCE_DATA',
        0x68: 'APCB_COPY',
        0x69: 'EARLY_VGA_IMAGE',
        0x6A: 'MP2_FW_CFG',
        0x80: 'OEM_System_Trusted_Application',
        0x81: 'OEM_System_TA_Signing_key',
        0x108: 'PSP_SMU_FN_FIRMWARE',
        0x118: 'PSP_SMU_FN_FIRMWARE2',

        # Entry types named by us
        #   Custom names are denoted by a leading '!'
        0x14: '!PSP_MCLF_TRUSTLETS',  # very similiar to ~PspTrustlets.bin~ in coreboot blobs
        0x40: '!PL2_SECONDARY_DIRECTORY',
        0x43: '!KEY_UNKNOWN_1',
        0x4e: '!KEY_UNKNOWN_2',
        0x70: '!BL2_SECONDARY_DIRECTORY',
        0x15f: '!FW_PSP_SMUSCS_2',  # seems to be a secondary FW_PSP_SMUSCS (see above)
        0x112: '!SMU_OFF_CHIP_FW_3',  # seems to tbe a tertiary SMU image (see above)
        0x10062: '!UEFI-IMAGE',
        0x30062: '!UEFI-IMAGE',
        0xdead: '!KEY_NOT_IN_DIR'

    }

    KEY_STORE_TYPES = [0x50, 0x51]

    class Type(Enum):
        NO_HDR_ENTRY = 1
        PUBKEY = 2
        NORMAL = 3

    class ParseError(Exception):
        pass

    class TypeError(Exception):
        pass

    @classmethod
    def from_fields(cls, parent_directory, parent_buffer, type_, size, offset, blob, psptool, destination: int = None):
        # Try to parse these ID's as a key entry
        # todo: consolidate these constants with Directory._ENTRY_TYPES_PUBKEY
        PUBKEY_ENTRY_TYPES = [0x0, 0x9, 0xa, 0x5, 0xd, 0x43, 0x4e, 0xdead]

        # Types known to have no PSP HDR
        # TODO: Find a better way to identify those entries
        NO_HDR_ENTRY_TYPES = [0x4, 0xb, 0x21, 0x40, 0x70, 0x30062, 0x6, 0x61, 0x60,
                              0x68, 0x100060, 0x100068, 0x5f, 0x15f, 0x1a, 0x22, 0x63,
                              0x67, 0x66, 0x100066, 0x200066, 0x300066, 0x10062,
                              0x400066, 0x500066, 0x800068, 0x61, 0x200060, 0x300060,
                              0x300068, 0x400068, 0x500068, 0x400060, 0x500060, 0x200068,
                              0x7, 0x38, 0x46, 0x54, 0x600060, 0x700060, 0x600068, 0x700068]

        NO_SIZE_ENTRY_TYPES = [0xb]

        size &= 0x00ffffff

        new_entry = None

        if type_ in NO_SIZE_ENTRY_TYPES:
            size = 0

        if type_ in NO_HDR_ENTRY_TYPES:
            # Option 1: it's a plain Entry
            try:
                new_entry = Entry(
                    parent_directory,
                    parent_buffer,
                    type_,
                    size,
                    offset,
                    blob,
                    psptool,
                    destination=destination,
                )
            except:
                psptool.ph.print_warning(f"Couldn't parse plain entry: 0x{type_:x}")

        elif type_ in PUBKEY_ENTRY_TYPES:
            # Option 2: it's a PubkeyEntry
            try:
                new_entry = PubkeyEntry(parent_directory, parent_buffer, type_, size, offset, blob, psptool)
            except Exception as e:
                new_entry = Entry(
                    parent_directory,
                    parent_buffer,
                    type_,
                    size,
                    offset,
                    blob,
                    psptool,
                    destination=destination,
                )
                psptool.ph.print_warning(f"{e.__class__.__name__} for {new_entry}")

        elif type_ in Entry.KEY_STORE_TYPES:
            # Option 2: it's a KeyStoreEntry
            try:
                new_entry = KeyStoreEntry(parent_directory, parent_buffer, type_, size, offset, blob, psptool)
            except:
                new_entry = Entry(
                    parent_directory,
                    parent_buffer,
                    type_,
                    size,
                    offset,
                    blob,
                    psptool,
                    destination=destination,
                )

        if new_entry is None:
            # Option 3: it's a HeaderEntry (most common)
            if size == 0:
                # If the size in the directory is zero, set the size to hdr len
                size = HeaderEntry.HEADER_LEN
            try:
                new_entry = HeaderEntry(parent_directory, parent_buffer, type_, size, offset, blob, psptool)
                if size == 0:
                    psptool.ph.print_warning(f"Entry with zero size. Type: {type_}. Dir: 0x{offset:x}")
            except:
                new_entry = Entry(
                    parent_directory,
                    parent_buffer,
                    type_,
                    size,
                    offset,
                    blob,
                    psptool,
                    destination=destination,
                )

        return new_entry

    @classmethod
    def from_blob(cls, binary, id_, type_, compressed, signed, psptool, private_key: PrivateKey=None):
        if type_ == Entry.Type.PUBKEY:
            psptool.ph.print_warning(f"from_blob is not implemented for pubkeys")
            pass
        elif type_ == Entry.Type.NO_HDR_ENTRY:
            psptool.ph.print_warning(f"from_blob is not implemented for non-header objects")
            pass
        elif type_ == Entry.Type.NORMAL:
            size = round_to_int(len(binary), 0x10)
            if compressed:
                rom_data = zlib_compress(binary)
                zlib_size = len(rom_data)
                padded_size = round_to_int(zlib_size, 0x10)
            else:
                rom_data = binary
                zlib_size = 0
                padded_size = round_to_int(len(rom_data), 0x10)

            if signed:
                assert private_key is not None
                total_size = padded_size + private_key.key_type.signature_size
            else:
                total_size = padded_size

            # Add 0x100 for the header
            total_size += 0x100

            if compressed:
                padding_size = padded_size - zlib_size
                total_size += padding_size
                blob = NestedBuffer(bytearray(total_size), total_size)
                blob[0x100:0x100 + zlib_size] = rom_data
                blob[0x100 + zlib_size:0x100 + padded_size] = padding_size * b'\xff'
            else:
                padding_size = padded_size - len(rom_data)
                total_size += padding_size
                blob = NestedBuffer(bytearray(total_size), total_size)
                blob[0x100:0x100 + len(rom_data)] = rom_data
                blob[0x100 + len(rom_data):0x100 + padded_size] = padded_size * b'\xff'

            # Set compressed bit
            if compressed:
                blob[0x48:0x4c] = (1).to_bytes(4, 'little')
            # Set size
            blob[0x14:0x18] = size.to_bytes(4, 'little')
            # Set rom_size
            blob[0x6c:0x70] = total_size.to_bytes(4, 'little')
            if compressed:
                # Set zlib_size
                blob[0x54:0x58] = zlib_size.to_bytes(4, 'little')

            entry = HeaderEntry(None, blob, id_, total_size, 0x0, blob, psptool)

            if signed:
                entry.signature[:] = private_key.sign(entry.get_signed_bytes())

            return entry
        else:
            raise Entry.TypeError()

    def __init__(self, parent_directory, parent_buffer, type_, buffer_size, buffer_offset: int, blob, psptool,
                 destination: int = None):
        super().__init__(parent_buffer, buffer_size, buffer_offset=buffer_offset)

        # TODO: Fix to reference of FET
        self.blob = blob
        self.psptool = psptool
        self.type = type_
        self.destination = destination
        # todo: deduplicate Entry objects pointing to the same address (in `from_fields`?)
        self.references = [parent_directory] if parent_directory is not None else []
        self.parent_directory = parent_directory

        self.compressed = False
        self.encrypted = False
        self.is_legacy = False
        self.sha256_verified = False

        try:
            self._parse()
        except (struct.error, AssertionError):
            self.psptool.ph.print_warning(f"Couldn't parse entry at: 0x{self.get_address():x}. "
                                          f"Type: {self.get_readable_type()}. Size 0x{len(self):x}")
            raise Entry.ParseError()

    @property
    def signed(self) -> bool:
        return False

    @property
    def has_sha256_checksum(self) -> bool:
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(type={hex(self.type)}, address={hex(self.get_address())}, ' \
               f'size={hex(self.buffer_size)}, len(references)={len(self.references)})'

    def __eq__(self, other):
        return self.type == other.type and self.get_address() == other.get_address() and \
              self.buffer_size == other.buffer_size

    def __hash__(self):
        return hash((self.type, self.get_address(), self.buffer_size))

    def __lt__(self, other):
        return self.get_address() < other.get_address()

    def _parse(self):
        pass

    def get_readable_type(self):
        if self.type in BIOS_ENTRY_TYPES:
            return "BIOS"
        if self.type in self.DIRECTORY_ENTRY_TYPES:
            return f'{self.DIRECTORY_ENTRY_TYPES[self.type]}~{hex(self.type)}'
        else:
            return hex(self.type)

    def get_readable_destination_address(self):
        return hex(self.destination)

    def get_readable_version(self):
        return ''

    def get_readable_magic(self):
        return ''

    def get_readable_signed_by(self):
        return ''

    def shannon_entropy(self):
        return shannon(self[:])

    def md5(self):
        m = md5()
        m.update(self.get_bytes())
        return m.hexdigest()

    def move_buffer(self, new_address, size):
        current_address = self.get_address()
        move_offset = new_address - current_address
        self.buffer_offset += move_offset
        self.buffer_size = int(ceil(size / self.ENTRY_ALIGNMENT)) * self.ENTRY_ALIGNMENT

        # update all directories' headers that point to this entry
        for directory in self.references:
            directory.update_entry_fields(self, self.type, self.buffer_size, self.buffer_offset)


class KeyStoreEntry(Entry):

    def _parse(self):

        # Will be set by the CertificateTree created after the blob
        self.signed_entity = None

        self.header = KeyStoreEntryHeader(self)

        key_store_start = self.header.buffer_size
        key_store_size = self.header.body_size
        self.key_store = KeyStore(self, key_store_size, buffer_offset=key_store_start)

        signature_start = key_store_start + key_store_size
        signature_size = self.header.signature_size
        assert signature_size in {0x100, 0x200}
        self.signature = Signature(self, signature_size, signature_start)
        #self.signature = ReversedSignature(self, signature_size, signature_start)

        if self.header.has_sha256_checksum:
            self.sha256_verified = self.verify_sha256()

        assert signature_start + signature_size == self.buffer_size

    def get_signed_bytes(self):
        return self.header.get_bytes() + self.key_store.get_bytes()

    def get_readable_version(self):
        return '1'

    def get_readable_magic(self):
        return f'{self.header.magic}'[2:-1]

    def get_readable_signed_by(self):
        return self.header.certifying_id.magic

    @property
    def signed(self):
        return True

    @property
    def _sha256_checksum_flag_2(self):
        nb = NestedBuffer(self, 0x4, 0x58)
        return struct.unpack(">I", nb.get_bytes())[0]

    @property
    def has_sha256_checksum(self) -> bool:
        return self.header.has_sha256_checksum

    def verify_sha256(self, print_warning=True) -> bool:
        if self.header.sha256_checksum.get_bytes() == sha256(self.key_store.get_bytes()).digest():
            return True
        if print_warning:
            self.psptool.ph.print_warning(f"Could not verify sha256 checksum for {self}")
        return False

    def update_sha256(self):
        self.header.sha256_checksum[:] = sha256(self.key_store.get_bytes()).digest()
        self.verify_sha256()


class KeyStoreEntryHeader(NestedBuffer):

    HEADER_SIZE = 0x100

    def __init__(self, entry):
        super().__init__(entry, self.HEADER_SIZE)

        self._magic = NestedBuffer(self, 0x4, buffer_offset=0x10)
        assert self.magic in {b'$PS1', 4*b'\0'}

        self._body_size = NestedBuffer(self, 0x4, buffer_offset=0x14)
        self._packed_size = NestedBuffer(self, 0x4, buffer_offset=0x6c)
        assert self.signature_size in {0x100, 0x200}

        self.certifying_id = KeyId(self, 0x10, buffer_offset=0x38)

        self._unknown_constant_1 = NestedBuffer(self, 0x4, buffer_offset=0x30)
        self._unknown_constant_2 = NestedBuffer(self, 0x4, buffer_offset=0x34)
        assert self.unknown_constants == (b'\1\0\0\0', b'\2\0\0\0')

        self._keystore_type = NestedBuffer(self, 0x4, buffer_offset=0x7c)
        assert self.keystore_type in Entry.KEY_STORE_TYPES or self.keystore_type == 0


        self._sha256_checksum_flag_1 = NestedBuffer(self, 0x4, buffer_offset=0x4c)
        self._sha256_checksum_flag_2 = NestedBuffer(self, 0x4, buffer_offset=0x58)

        self.sha256_checksum = None
        if self.has_sha256_checksum:
            self.sha256_checksum = NestedBuffer(self, 0x20, buffer_offset=0xd0)

        zero_ranges = {
            (0x00, 0x10),
            (0x18, 0x18),
            (0x48, 0x04),
            (0x50, 0x08),
            (0x5c, 0x10),
            (0x70, 0x0c),
            (0x80, 0x50),
            (0xf0, 0x10),
        }
        for (start, length) in zero_ranges:
            assert self.get_bytes(start, length) == b'\0' * length

    @property
    def magic(self) -> bytes:
        return self._magic.get_bytes()

    @magic.setter
    def magic(self, value: bytes):
        self._magic[:] = value

    @property
    def body_size(self) -> int:
        return int.from_bytes(self._body_size.get_bytes(), 'little')

    @property
    def packed_size(self) -> int:
        return int.from_bytes(self._packed_size.get_bytes(), 'little')

    @property
    def keystore_type(self) -> int:
        return int.from_bytes(self._keystore_type.get_bytes(), 'little')

    @property
    def signature_size(self) -> int:
        return self.packed_size - self.HEADER_SIZE - self.body_size

    @property
    def unknown_constants(self) -> (bytes, bytes, bytes):
        return (
                self._unknown_constant_1.get_bytes(),
                self._unknown_constant_2.get_bytes(),
                )

    @property
    def sha256_checksum_flag_1(self) -> int:
        return int.from_bytes(self._sha256_checksum_flag_1.get_bytes(), 'little')

    @property
    def sha256_checksum_flag_2(self) -> int:
        return int.from_bytes(self._sha256_checksum_flag_2.get_bytes(), 'big')

    @property
    def has_sha256_checksum(self) -> bool:
        # assert self.sha256_checksum_flag_1 == self.sha256_checksum_flag_2
        assert self.sha256_checksum_flag_1 in {0, 1}
        return self.sha256_checksum_flag_1 == 1


class KeyStore(NestedBuffer):

    HEADER_SIZE = 0x50

    def __init__(self, parent_buffer, buffer_size: int, buffer_offset: int = 0):
        super().__init__(parent_buffer, buffer_size, buffer_offset)

        # parse header
        self.header = NestedBuffer(self, self.HEADER_SIZE)

        self._size = NestedBuffer(self.header, 0x4, buffer_offset=0)
        assert self.size == self.buffer_size

        self._unknown_flag = NestedBuffer(self.header, 0x4, buffer_offset=0x4)
        assert self.unknown_flag

        self.magic = NestedBuffer(self.header, 0x4, buffer_offset=0x8)
        assert self.magic.get_bytes() == b'$KDB'

        assert self.header.get_bytes(0xc, 0x44) == b'\0' * 0x44

        # parse body
        body_start = self.header.buffer_size
        body_size = self.buffer_size - body_start
        assert body_size > 0
        self.body = NestedBuffer(self, body_size, buffer_offset=body_start)

        next_key_start = 0
        self.keys = []
        while next_key_start < body_size:
            key = KeyStoreKey(self.body, next_key_start)
            self.keys.append(key)
            next_key_start += key.size

    @property
    def size(self) -> int:
        return int.from_bytes(self._size.get_bytes(), 'little')

    @property
    def unknown_flag(self) -> bool:
        value = int.from_bytes(self._unknown_flag.get_bytes(), 'little')
        assert value in {0,1}
        return value == 1


class KeyStoreKey(NestedBuffer):

    HEADER_SIZE = 0x50

    def __init__(self, body: NestedBuffer, offset: int):

        # Will be set by the CertificateTree created after the blob
        self.pubkey_entity = None

        # init self
        size = int.from_bytes(body.get_bytes(offset, 0x4), 'little')
        super().__init__(body, size, buffer_offset=offset)

        # init header
        self.header = NestedBuffer(body, self.HEADER_SIZE, buffer_offset=offset)

        # init crypto_material (body)
        body_start = offset + self.HEADER_SIZE
        body_size = size - self.HEADER_SIZE
        assert body_size > 0
        self.crypto_material = NestedBuffer(body, body_size, buffer_offset=body_start)

        # init header fields
        self._size = NestedBuffer(self.header, 0x4, buffer_offset=0)
        assert self.size == self.buffer_size

        self._unknown_flag = NestedBuffer(self.header, 0x4, buffer_offset=0x4)
        assert self.unknown_flag

        self._unknown_id = NestedBuffer(self.header, 0x4, buffer_offset=0x8)
        assert self.unknown_id < 0x100

        self._rsa_exponent = NestedBuffer(self.header, 0x4, buffer_offset=0xc)
        assert self.rsa_exponent == 0x10001

        self.key_id = KeyId(self.header, 0x10, buffer_offset=0x10)

        self._key_size = NestedBuffer(self.header, 0x4, buffer_offset=0x20)
        assert self.key_size == self.crypto_material.buffer_size << 3

        assert self.header.get_bytes(0x24, 0x2c) == b'\0' * 0x2c

    @property
    def size(self) -> int:
        return int.from_bytes(self._size.get_bytes(), 'little')

    @property
    def unknown_flag(self) -> bool:
        value = int.from_bytes(self._unknown_flag.get_bytes(), 'little')
        assert value in {0,1}
        return value == 1

    @property
    def unknown_id(self) -> int:
        return int.from_bytes(self._unknown_id.get_bytes(), 'little')

    @property
    def rsa_exponent(self) -> int:
        return int.from_bytes(self._rsa_exponent.get_bytes(), 'little')

    @property
    def key_size(self) -> int:
        return int.from_bytes(self._key_size.get_bytes(), 'little')


class UnknownPubkeyEntryVersion(Exception):
    pass


class PubkeyEntry(Entry):

    HEADER_LEN = 0x40

    def _parse(self):
        """ SEV spec B.1 """

        # Will be set by the CertificateTree created after the blob
        self.signed_entity = None
        self.pubkey_entity = None

        # Will be set by blob.find_inline_pubkeys
        self.is_inline = False
        self.parent_entry = None

        # misc info
        self._version = NestedBuffer(self, 4)
        if self.version not in {1, 2}:
            raise UnknownPubkeyEntryVersion
        self._key_usage = NestedBuffer(self, 4, 0x24)

        # key ids
        self.key_id = KeyId(self, 0x10, 0x4)
        self.certifying_id = KeyId(self, 0x10, 0x14)

        # security features
        self._security_features = NestedBuffer(self, 2, 0x2A)

        # crypto material
        self._pubexp_bits = NestedBuffer(self, 4, 0x38)
        self._modulus_bits = NestedBuffer(self, 4, 0x3c)
        assert self.pubexp_bits == self.modulus_bits
        assert self.pubexp_bits in {2048, 4096}

        self.crypto_material = NestedBuffer(self, self.pubexp_size + self.modulus_size, self.HEADER_LEN)
        self._pubexp = NestedBuffer(self.crypto_material, self.pubexp_size)
        self._modulus = NestedBuffer(self.crypto_material, self.modulus_size, self.pubexp_size)
        assert self.pubexp == 0x10001

        # signature
        if self.signed:
            assert self.signature_size in {0x100, 0x200}
            signature_start = self.HEADER_LEN + self.pubexp_size + self.modulus_size
            self.signature = ReversedSignature(self, self.signature_size, signature_start)

    @property
    def version(self) -> int:
        return int.from_bytes(self._version.get_bytes(), 'little')

    @property
    def key_usage(self) -> int:
        return int.from_bytes(self._key_usage.get_bytes(), 'little')

    @property
    def pubexp_bits(self) -> int:
        return int.from_bytes(self._pubexp_bits.get_bytes(), 'little')

    @property
    def modulus_bits(self) -> int:
        return int.from_bytes(self._modulus_bits.get_bytes(), 'little')

    @property
    def pubexp_size(self) -> int:
        assert self.pubexp_bits & 0x3 == 0
        return self.pubexp_bits >> 3

    @property
    def modulus_size(self) -> int:
        assert self.modulus_bits & 0x3 == 0
        return self.modulus_bits >> 3

    @property
    def signature_size(self) -> int:
        return self.buffer_size - self.HEADER_LEN - self.pubexp_size - self.modulus_size

    @property
    def signed(self) -> bool:
        return self.signature_size != 0

    @property
    def pubexp(self) -> int:
        return int.from_bytes(self._pubexp.get_bytes(), 'little')

    @property
    def modulus(self) -> int:
        return int.from_bytes(self._modulus.get_bytes(), 'little')

    @property
    def security_features(self) -> int:
        return int.from_bytes(self._security_features.get_bytes(), 'little')

    def get_signed_bytes(self):
        return self.get_bytes(0, self.buffer_size - self.signature_size)

    def get_readable_signed_by(self):
        if self.signed:
            return self.certifying_id.magic

    def get_readable_magic(self):
        return self.key_id.magic

    def get_readable_version(self):
        return str(self.version)

    def get_readable_key_usage(self):
        if self.key_usage == 0:
            return 'AMD_CODE_SIGN'
        if self.key_usage == 1:
            return 'BIOS_CODE_SIGN'
        if self.key_usage == 2:
            return 'AMD_AND_BIOS_CODE_SIGN'
        if self.key_usage == 8:
            return 'PLATFORM_SECURE_BOOT'
        return f'unknown_key_usage({self.key_usage})'

    def get_readable_security_features(self):
        features = []
        if self.security_features & 0b001:
            features.append('DISABLE_BIOS_KEY_ANTI_ROLLBACK')
        if self.security_features & 0b010:
            features.append('DISABLE_AMD_BIOS_KEY_USE')
        if self.security_features & 0b100:
            features.append('DISABLE_SECURE_DEBUG_UNLOCK')
        return ', '.join(features)

class HeaderEntry(Entry):

    HEADER_LEN = 0x100

    def _parse(self):
        self.header = NestedBuffer(self, HeaderEntry.HEADER_LEN)

        # Will be set by the CertificateTree created after the blob
        self.signed_entity = None

        # Will be set by blob._find_inline_pubkeys
        self.inline_keys = set()

        # todo: use NestedBuffers instead of saving by value
        self.magic = self.header[0x10:0x14]
        self.size_signed = struct.unpack('<I', self.header[0x14:0x18])[0]
        self.encrypted = struct.unpack('<I', self.header[0x18:0x1c])[0] == 1
        self._signed = NestedBuffer(self, 4, 0x30)
        self.signature_fingerprint = hexlify(self.header[0x38:0x48])
        self.compressed = struct.unpack('<I', self.header[0x48:0x4c])[0] == 1
        self.size_uncompressed = struct.unpack('<I', self.header[0x50:0x54])[0]
        self.version = self.header[0x63:0x5f:-1]
        self.load_addr = struct.unpack('<I', self.header[0x68:0x6c])[0]
        self.rom_size = struct.unpack('<I', self.header[0x6c:0x70])[0]
        self.zlib_size = struct.unpack('<I', self.header[0x54:0x58])[0]

        self.iv = hexlify(self.header[0x20:0x30])
        self.unknown_bool = struct.unpack('<I', self.header[0x7c:0x80])[0]
        self.wrapped_key = hexlify(self.header[0x80:0x90])

        # TODO: Take care of headers with only 0xfff...
        # TODO if zlib_size == 0 try size_signed

        assert(self.compressed in [0, 1])
        assert(self.encrypted in [0, 1])

        if self.signed:
            self._parse_signature()
        else:
            self.signature_len = 0

        self.header_len = 0x100

        if self.rom_size == 0 or (self.compressed and self.zlib_size == 0):
            # Try to parse as legacy header
            self._parse_legacy_hdr()
        else:
            self._parse_hdr()

        self._sha256_checksum = NestedBuffer(self, 0x20, 0xd0)
        if self.has_sha256_checksum:
            self.sha256_verified = self.verify_sha256()

        return

    def _parse_signature(self):
        if self.signature_fingerprint != hexlify(16 * b'\x00'):

            body_size = self.size_signed
            if self.compressed:
                body_size = self.zlib_size

            self.signature_len = self.rom_size - 0x100 - body_size
            if self.signature_len < 0:
                self.signature_len = 0

            # Round to 0x100, 0x200, etc.
            self.signature_len >>= 8
            self.signature_len <<= 8

            if self.signature_len > 0x200:

                # this is a best-effort guess made for e.g. PSP_FW_TRUSTED_OS~0x2
                self.signature_len = 0x100

            if self.signature_len % 0x100 > 0x10:
                # self.psptool.ph.print_warning(f"Signature size of 0x{self.signature_len:x} seems odd!")
                pass

            if self.signature_len not in {0x100, 0x200}:
                # self.psptool.ph.print_warning(f"Signature size of 0x{self.signature_len:x} seems odd!")
                # self.psptool.ph.print_warning(f"signe_sz=0x{self.size_signed:x}")
                # self.psptool.ph.print_warning(f"rom_sz=0x{self.rom_size:x}")
                # self.psptool.ph.print_warning(f"zlib_sz=0x{self.zlib_size:x}")
                pass

            # self.psptool.ph.print_warning(f"Couldn't find corresponding key in blob for entry at: 0x{self.get_address():x}. Type: "
                              # f"{self.get_readable_type()}")
        else:
            self.psptool.ph.print_warning("ERROR: Signed but no key id present")

    def _parse_legacy_hdr(self):
        self.buffer_size = self.size_signed + self.header_len + self.signature_len
        self.buffer_size &= 0x00ffffff

        if self.compressed:
            self.zlib_size = self.size_signed

        if self.signed:
            self.signature = NestedBuffer(self, self.signature_len, self.buffer_size - self.signature_len)

        self.body = NestedBuffer(self, len(self) - self.header_len, self.header_len)

        self.is_legacy = True

    def _parse_hdr(self):
        if self.rom_size == 0:
            # TODO throw exception
            self.buffer_size = self.size_signed + self.header_len
            self.psptool.ph.print_warning("ERROR. rom size is zero")
        else:
            self.buffer_size = self.rom_size

        if self.signed:
            buf_start = self.get_address()
            sig_start = self.get_address() + self.rom_size - self.signature_len
            # self.psptool.ph.print_warning(f"Signature at: 0x{buf_start:x} sig_start: 0x{sig_start:x}")
            self.signature = NestedBuffer(self, self.signature_len, sig_start - buf_start)

        if self.compressed:
            if self.zlib_size == 0:
                # Todo throw exception
                self.psptool.ph.print_warning(f"ERROR: Weird entry. Address 0x{self.get_address():x}")

        # Get IV and wrapped KEY from entry header
        if self.encrypted:
            self.iv = self.header[0x20:0x30]
            self.key = self.header[0x80:0x90]
            assert(self.iv != (b'\x00' * 16))
            assert(self.key != (b'\x00' * 16))

        self.body = NestedBuffer(self, len(self) - self.header_len - self.signature_len, self.header_len)
        self.is_legacy = False

    @property
    def signed(self) -> bool:
        signed = int.from_bytes(self._signed.get_bytes(), 'little')
        assert signed in {0, 1, 0xffff0000}, f'did not expect signed to be 0x{signed:x}'
        return signed != 0

    # @property
    # def _sha256_checksum_flag_1(self):
    #     nb = NestedBuffer(self, 0x4, 0x4c)
    #     return struct.unpack(">I", nb.get_bytes())[0]
    #
    # @_sha256_checksum_flag_1.setter
    # def _sha256_checksum_flag_1(self, value):
    #     nb = NestedBuffer(self, 0x4, 0x4c)
    #     nb[:] = value

    @property
    def _sha256_checksum_flag_2(self):
        nb = NestedBuffer(self, 0x4, 0x58)
        return struct.unpack(">I", nb.get_bytes())[0]

    @property
    def has_sha256_checksum(self) -> bool:
        return self._sha256_checksum_flag_2 == 1

    def verify_sha256(self, print_warning=True) -> bool:
        if self._sha256_checksum.get_bytes() == sha256(self.get_decrypted_decompressed_body()).digest():
            return True
        if print_warning:
            self.psptool.ph.print_warning(f"Could not verify sha256 checksum for {self}")
        return False

    def update_sha256(self):
        self._sha256_checksum[:] = sha256(self.get_decrypted_decompressed_body()).digest()
        self.verify_sha256()

    def get_readable_version(self):
        return '.'.join([hex(b)[2:].upper() for b in self.version])

    def get_ikek_md5sum(self) -> bytes:
        ikek = self.parent_buffer.get_entries_by_type(0x21)[0]
        m = md5()
        m.update(ikek.get_bytes())
        return m.digest()

    def get_readable_magic(self):
        # if self.magic == b'\x01\x00\x00\x00':
            # actually twice as long, but SMURULESMURULES is kinda redundant
            # readable_magic= self[0x0:0x4]
        if self.magic == b'\x05\x00\x00\x00':
            readable_magic = b'0x05'
        else:
            readable_magic = self.magic

        try:
            # Try to encode the id as ascii
            readable_magic = str(readable_magic, encoding='ascii')
            # and remove unprintable chars
            readable_magic = ''.join(s for s in readable_magic if s in string.printable)
        except UnicodeDecodeError:
            return ''

        return readable_magic

    def get_readable_signed_by(self):
        return self.signed_entity.certifying_id.magic

    def get_signed_bytes(self) -> bytes:
        entry_bytes = self.header.get_bytes() + self.get_decrypted_decompressed_body()
        return entry_bytes[:self.header_len + self.size_signed]

    def get_decrypted_decompressed_body(self) -> bytes:
        if self.encrypted:
            output = self.get_decrypted_body()
        else:
            output = self.body.get_bytes()
        if self.compressed:
            try:
                return zlib_decompress(output[:self.zlib_size])
            except:
                self.psptool.ph.print_warning(f"ZLIB decompression failed on entry {self.get_readable_type()}")
        return output

    def to_decrypted_entry_bytes(self) -> bytes:
        """Returns the bytes of the same entry, just with the encryption removed"""
        header = bytearray(self.header.get_bytes())
        header[0x18:0x1c] = bytes(4)
        header[0x20:0x30] = bytes(0x10)
        signature = self.signature.get_bytes() if self.signed else b''
        return bytes(header) + self.get_decrypted_body() + signature

    def get_decrypted_body(self) -> bytes:
        if not self.encrypted:
            return self.body.get_bytes()
        else:
            unwrapped_ikek = self.get_unwrapped_ikek()
            assert(unwrapped_ikek != None)
            return decrypt(self.body.get_bytes(), self.key, unwrapped_ikek, self.iv)

    def get_unwrapped_ikek(self) -> bytes:
        # TODO: Find out how to identify the correct IKEK.
        #       For now assume that the zen+ IKEK is correct.

        # if self.get_ikek_md5sum() == self.HASH_IKEK_ZEN:
        #     return self.UNWRAPPED_IKEK_ZEN
        # if self.get_ikek_md5sum() == self.HASH_IKEK_ZEN_PLUS:
        #     return self.UNWRAPPED_IKEK_ZEN_PLUS
        # else:
        #     return None

        return self.UNWRAPPED_IKEK_ZEN_PLUS

    def shannon_entropy(self):
        return shannon(self.body[:])

    def md5(self):
        m = md5()
        try:
            m.update(self.body.get_bytes())
        except:
            self.psptool.ph.print_warning(f"Get bytes failed at entry: 0x{self.get_address():x} type: {self.get_readable_type()} size: 0x{self.buffer_size:x}")
        return m.hexdigest()




# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .utils import NestedBuffer


class Firmware(NestedBuffer):
    # todo: find out correct size
    def __init__(self, parent_buffer, buffer_offset: int, firmware_type: str, magic: bytes):
        super().__init__(parent_buffer, 0x100, buffer_offset=buffer_offset)

        self.magic = magic
        self.type = firmware_type

    def __repr__(self):
        return f'<Firmware(type={self.type}, address={hex(self.get_address())}, magic={self.magic})>)'



# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from prettytable import PrettyTable
import sys, json

from .entry import Entry, HeaderEntry, PubkeyEntry
from .blob import Blob
from .utils import PrintHelper
from .cert_tree import CertificateTree
from . import errors


class PSPTool:
    @classmethod
    def from_file(cls, filename, verbose=False):
        with open(filename, 'rb') as f:
            file_bytes = bytearray(f.read())

        pt = PSPTool(file_bytes, verbose=verbose)
        pt.filename = filename

        return pt

    def __init__(self, rom_bytes, verbose=False):
        self.ph = PrintHelper(verbose)

        self.blob = Blob(rom_bytes, len(rom_bytes), self)
        self.cert_tree = CertificateTree.from_blob(self.blob, self)

        self.filename = None

    def __repr__(self):
        if self.filename is not None:
            return f'PSPTool(filename={self.filename})'
        else:
            return f'PSPTool(len(rom_bytes)={self.blob.buffer_size}'

    def to_file(self, filename):
        with open(filename, 'wb') as f:
            f.write(self.blob.get_buffer())

    def to_stdout(self):
        sys.stdout.buffer.write(self.blob.get_buffer())

    def ls(self, verbose=False):
        for rom_index, rom in enumerate(self.blob.roms):
            t = PrettyTable(['ROM', 'Addr', 'Size', 'FET', 'AGESA'])
            t.add_row([
                rom_index,
                hex(rom.get_address()),
                hex(rom.buffer_size),
                hex(rom.fet.get_address()),
                rom.agesa_version
            ])
            print(t)

            for index, directory in enumerate(rom.directories):
                t = PrettyTable(['', 'Directory', 'Addr', 'Type', 'Generation', 'Magic', 'Secondary Directory'])
                t.add_row([
                    '',
                    index,
                    hex(directory.get_address()),
                    directory.type,
                    directory.zen_generation,
                    directory.magic.decode('utf-8', 'backslashreplace'),
                    hex(directory.secondary_directory_address) if directory.secondary_directory_address else '--'
                ])

                print(t)

                self.ls_dir(rom, index, verbose=verbose)
                print('\n')

    def ls_dir(self, fet,  directory_index, verbose=False):
        directory = fet.directories[directory_index]
        self.ls_entries(entries=directory.entries, verbose=verbose)

    def ls_entries(self, entries=None, verbose=False):
        # list all entries of all directories by default (sorted by their address)
        if entries is None:
            entries = sorted(self.blob.unique_entries())

        basic_fields = ['', ' ', 'Entry', 'Address', 'Size', 'Type', 'Magic/ID', 'Version', 'Info']
        verbose_fields = ['MD5', 'size_signed', 'size_full', 'size_packed', 'load_addr']

        t = PrettyTable(basic_fields + verbose_fields)
        t.align = 'r'

        # TODO: Skip this whole mess and introduce strict and non_strict mode
        #  strict mode should parse everything but give inconsistency errors like sha256_inconsistent
        entry: Entry
        for index, entry in enumerate(entries):
            info = []
            if entry.compressed:
                info.append('compressed')
            if entry.signed:
                try:
                    if entry.signed_entity.is_verified():
                        info.append(f'verified({entry.get_readable_signed_by()})')
                    else:
                        info.append(f'veri-failed({entry.get_readable_signed_by()})')
                except errors.NoCertifyingKey:
                    info.append(f'key_missing({entry.signed_entity.certifying_id.as_string()[:4]})')
                except errors.SignatureInvalid:
                    info.append(f'invalid_sig({entry.get_readable_signed_by()})')
            if entry.has_sha256_checksum:
                if entry.sha256_verified:
                    info.append(f'sha256_ok')
                else:
                    info.append(f'sha256_inconsistent')
            if entry.is_legacy:
                info.append('legacy_header')
            if entry.encrypted:
                info.append('encrypted')
            if type(entry) == HeaderEntry and entry.inline_keys:
                inline_keys = ', '.join(map(lambda k: k.get_readable_magic(), entry.inline_keys))
                info.append(f'inline_keys({inline_keys})')
            if type(entry) == PubkeyEntry:
                info.append(entry.get_readable_key_usage())
                if entry.get_readable_security_features():
                    info.append(entry.get_readable_security_features())

            all_values = [
                '',
                '',
                index,
                hex(entry.get_address()),
                hex(entry.buffer_size),
                entry.get_readable_type(),
                entry.get_readable_magic(),
                entry.get_readable_version(),
                ', '.join(info),
                entry.md5()[:4].upper()
            ]

            if type(entry) is HeaderEntry:
                all_values += [hex(v) for v in [
                    entry.size_signed,
                    entry.size_uncompressed,
                    entry.rom_size,
                    entry.load_addr
                ]]
            else:
                all_values += (4 * [''])

            t.add_row(all_values)

        fields = basic_fields

        if verbose is True:
            fields += verbose_fields

        print(t.get_string(fields=fields))

    def ls_json(self, verbose=False):
        data = []
        # todo: add notion of Multi-ROMs
        for rom in self.blob.roms:
            for index, directory in enumerate(rom.directories):
                PrettyTable(['Directory', 'Addr', 'Type', 'Magic', 'Secondary Directory'])
                d = {
                    'directory': index,
                    'address': directory.get_address(),
                    'directoryType': directory.type,
                    'magic': directory.magic.decode('utf-8', 'backslashreplace'),
                }
                if directory.secondary_directory_address:
                    d['secondaryAddress'] = directory.secondary_directory_address

                entries = self.ls_dir_dict(rom, index, verbose=verbose)
                d['entries'] = entries
                data.append(d)
        print(json.dumps(data))

    def ls_dir_dict(self, fet,  directory_index, verbose=False):
        directory = fet.directories[directory_index]
        return self.ls_entries_dict(entries=directory.entries)

    def ls_entries_dict(self, entries=None):
        # list all entries of all directories by default (sorted by their address)
        if entries is None:
            entries = sorted(self.rom.unique_entries)

        out = []
        for index, entry in enumerate(entries):
            info = []
            if entry.compressed:
                info.append('compressed')
            if entry.signed:
                info.append(f'signed({entry.get_readable_signed_by()})')
                try:
                    if entry.signed_entity.is_verified():
                        info.append('verified')
                except errors.NoCertifyingKey:
                    info.append('no_key')
            if entry.is_legacy:
                info.append('legacy header')
            if entry.encrypted:
                info.append('encrypted')

            all_values = {
                'index': index,
                'address': entry.get_address(),
                'size': entry.buffer_size,
                'sectionType': entry.get_readable_type(),
                'magic': entry.get_readable_magic(),
                'version': entry.get_readable_version(),
                'info': info,
                'md5': entry.md5()[:4].upper()
            }

            if entry.get_readable_type() == "BIOS":
                all_values['destinationAddress'] = entry.get_readable_destination_address()

            if type(entry) is HeaderEntry:
                sizes = {
                    'signed': entry.size_signed,
                    'uncompressed': entry.size_uncompressed,
                    'packed': entry.rom_size
                }
                all_values['sizes'] = sizes

            out.append(all_values)

        return out



# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import argparse
import math
import zlib
import struct

from typing import Set, TypeVar

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend


class ObligingArgumentParser(argparse.ArgumentParser):
    """ Display the full help message whenever there is something wrong with the arguments.
        (from https://groups.google.com/d/msg/argparse-users/LazV_tEQvQw/xJhBOm1qS5IJ) """

    def error(self, message):
        sys.stderr.write('Error: %s\n' % message)
        self.print_help()
        sys.exit(2)


class NestedBuffer:
    def __init__(self, parent_buffer, buffer_size: int, buffer_offset: int = 0):
        self.parent_buffer = parent_buffer
        self.buffer_size = buffer_size
        self.buffer_offset = buffer_offset
        if hasattr(parent_buffer, 'buffer_size'):
            assert self.buffer_offset + self.buffer_size <= parent_buffer.buffer_size, \
                f'Cannot create child buffer: overflows parent buffer\'s bounds!'
        assert (self.buffer_size <= self.buffer_offset + self.buffer_size), \
            f'sz=0x{self.buffer_size:x},off=0x{self.buffer_offset:x}'

    def __len__(self):
        return self.buffer_size

    def __getitem__(self, item):
        if isinstance(item, slice):
            new_slice = self._offset_slice(item)
            return self.parent_buffer[new_slice]
        else:
            assert (isinstance(item, int))
            return self.parent_buffer[self.buffer_offset + item]

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            new_slice = self._offset_slice(key)
            self.parent_buffer[new_slice] = value
        else:
            assert (isinstance(key, int))
            self.parent_buffer[self.buffer_offset + key] = value

    def __repr__(self):
        return f'{self.__class__.__name__}(address={hex(self.get_address())}, buffer_size={hex(self.buffer_size)}, ' \
                f'buffer_offset={hex(self.buffer_offset)}, self.parent_buffer=0x{self.parent_buffer.get_address():x})'

    def _offset_slice(self, old_slice):
        if old_slice.start is None:
            start = self.buffer_offset
        else:
            assert (old_slice.start <= self.buffer_size), "Slice offset out of bounds"
            if old_slice.start < 0:
                start = self.buffer_offset + old_slice.start % self.buffer_size
            else:
                start = self.buffer_offset + old_slice.start
        if old_slice.stop is None:
            stop = self.buffer_offset + self.buffer_size
        else:
            assert (old_slice.stop <= self.buffer_size), \
                f'{old_slice.stop=},{self.buffer_size=}'
            if old_slice.stop < 0:
                stop = self.buffer_offset + old_slice.stop % self.buffer_size
            else:
                stop = self.buffer_offset + old_slice.stop

        new_slice = slice(start, stop, old_slice.step)
        return new_slice

    def get_address(self) -> int:
        if isinstance(self.parent_buffer, NestedBuffer):
            return self.buffer_offset + self.parent_buffer.get_address()
        else:
            return self.buffer_offset

    def get_buffer(self):
        return self.parent_buffer

    def get_bytes(self, offset: int = 0x0, size: int = None) -> bytes:
        size = self.buffer_size if size is None else size
        return bytes(self[offset:offset + size])

    def set_bytes(self, address: int, size: int, value):
        assert len(value) == size, "set_bytes: value length does not match size argument"
        self[address:address + size] = value

    def get_chunks(self, size: int, offset: int = 0):
        return chunker(self[offset:], size)


# from https://stackoverflow.com/a/39358140
class RangeDict(dict):
    def __getitem__(self, item):
        if type(item) != range:
            for key in self:
                if item in key:
                    return self[key]
        else:
            return super().__getitem__(item)


class PrintHelper:
    def __init__(self, is_verbose):
        self.is_verbose = is_verbose

    def print_error_and_exit(self, arg0):
        """ Wrapper function to print errors to stderr, so we don't interfere with e.g. extraction output. """
        arg0 = 'Error: ' + arg0 + '\n'
        sys.stderr.write(arg0)
        sys.exit(1)

    def print_warning(self, arg0):
        """ Wrapper function to print warnings to stderr, so we don't interfere with e.g. extraction output. """
        arg0 = 'Warning: ' + arg0 + '\n'
        sys.stderr.write(arg0)

    def print_info(self, arg0):
        """ Wrapper function to print info to stderr, so we don't interfere with e.g. extraction output. """
        # if self.is_verbose:
        arg0 = 'Info: ' + arg0 + '\n'
        sys.stderr.write(arg0)


def round_to_int(n, i):
    if n % i:
        return n + (i - n % i)
    else:
        return n


def chunker(seq, size):
    """ Utility function to chunk seq into a list of size sized sequences. """
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


def rstrip_padding(bytestring):
    """ Takes a bytestring and strips trailing 0xFFFFFFFF dwords. """
    i = 0
    while bytestring[-(4 + i):len(bytestring) - i] == b'\xff\xff\xff\xff':
        i += 4
    return bytestring[:len(bytestring) - i]


def shannon(s):
    """ Performs a Shannon entropy analysis on a given block of s.
    from: https://github.com/ReFirmLabs/binwalk """

    entropy = 0

    if s:
        length = len(s)

        seen = dict((x, 0) for x in range(0, 256))
        for byte in s:
            seen[byte] += 1

        for x in range(0, 256):
            p_x = float(seen[x]) / length
            if p_x > 0:
                entropy -= p_x * math.log(p_x, 2)

    return entropy / 8


# The order is important here, as 78da is the most common magic and others might produce false positives
ZLIB_TYPES = {
    b'\x78\xda': 'Zlib compressed data, best compression',
    b'\x78\x9c': 'Zlib compressed data, default compression',
    b'\x78\x5e': 'Zlib compressed data, compressed',
    b'\x78\x01': 'Zlib header, no compression'
}


def zlib_find_header(s):
    """ Checks s for any zlib magic bytes and returns the offset (or -1). """

    # Only check the first 0x500 bytes, as the rest is too unlikely to be valid
    s = s[:0x500]

    for zlib_magic in ZLIB_TYPES.keys():
        # Check the most common location at 0x100 first to avoid false positives and save time
        if s[0x100:0x102] == zlib_magic:
            return 0x100

        zlib_start = s.find(zlib_magic)

        if zlib_start != -1:
            return zlib_start

    return -1


def zlib_compress(s):
    return zlib.compress(s, 9)


def zlib_decompress(s):
    """ Checks s for the first appearance of a zlib header and returns the uncompressed start of s as well as the
    decompressed section. If no zlib header is found, s is returned as is. """

    zlib_start = zlib_find_header(s)

    if zlib_start != -1:
        uncompressed = s[:zlib_start]
        compressed = s[zlib_start:]
        decompressed = zlib.decompress(compressed)

        return uncompressed + decompressed

    return s


def decrypt_ecb(data, key):
    """ Decrypts 'data' with 'key' using AES-128 in ECB mode.
    Return the decrypted data. """
    backend = default_backend()

    cipher_ecb = Cipher(algorithms.AES(key), modes.ECB(), backend=backend)
    ecb_decryptor = cipher_ecb.decryptor()
    return ecb_decryptor.update(data) + ecb_decryptor.finalize()


def decrypt_cbd(data, iv, key):
    """ Decrypts 'data' with 'key' using AES-128 in CBC mode.
    Returns the decrypted data. """

    backend = default_backend()
    cipher_cbc = Cipher(algorithms.AES(key), modes.CBC(iv), backend=backend)
    cbc_decryptor = cipher_cbc.decryptor()
    return cbc_decryptor.update(data) + cbc_decryptor.finalize()


def decrypt(data, entry_key, unwrapped_ikek, iv):
    """ Decrypts an entry. The entry key is stored at offset 0x80 of the respective header,
    the IV is stored at offset 0x20. An already unwrapped IKEK is required to perform this
    operation. """
    unwrapped_entry_key = decrypt_ecb(entry_key, unwrapped_ikek)

    return decrypt_cbd(data, iv, unwrapped_entry_key)


def fletcher32(s):
    c0 = 0xFFFF
    c1 = 0xFFFF

    for index, byte_pair in enumerate(chunker(s, 2)):  # fletcher is calculated over 16bit words, i.e. 2 bytes
        byte_pair_int = struct.unpack('<H', byte_pair)[0]

        c0 += byte_pair_int
        c1 += c0

        if index % 360 == 0:
            c0 = (c0 & 0xFFFF) + (c0 >> 16)
            c1 = (c1 & 0xFFFF) + (c1 >> 16)

    c0 = (c0 & 0xFFFF) + (c0 >> 16)
    c1 = (c1 & 0xFFFF) + (c1 >> 16)

    c0 = (c0 & 0xFFFF) + (c0 >> 16)
    c1 = (c1 & 0xFFFF) + (c1 >> 16)

    checksum = (c1 << 16) | c0
    return struct.pack('<I', checksum)


T = TypeVar('T')


def sole(set_of_one: Set[T], assert_msg="Set does not contain exactly one element") -> T:
    assert len(set_of_one) == 1, assert_msg
    return list(set_of_one)[0]



# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import binascii

from typing import List

from .fet import EmptyFet
from .rom import Rom
from .utils import NestedBuffer, RangeDict
from .entry import Entry, PubkeyEntry


class Blob(NestedBuffer):
    _FIRMWARE_ENTRY_MAGIC = b'\xAA\x55\xAA\x55'
    _FIRMWARE_ENTRY_TABLE_BASE_ADDRESS = 0x20000

    class NoFirmwareEntryTableError(Exception):
        pass

    def __init__(self, buffer: bytearray, size: int, psptool):
        super().__init__(buffer, size)

        self.psptool = psptool
        self.roms: List[Rom] = []

        potential_fet_offsets = [
            # as seen by a PSPTrace Zen 1 boot
            0x020000,
            0xfa0000,
            0xf20000,
            0xe20000,
            0xc20000,
            0x820000,
        ]

        rom_size = 0x1000000
        if self.buffer_size < rom_size:
            self.psptool.ph.print_warning("Input  file < 16M, will assume 8M ROM ...")
            rom_size = 0x800000

        # For each FET, we try to create a 16MB ROM starting at `FET - offset`
        for fet_location in self._find_fets():
            fet_parsed = False
            for fet_offset in potential_fet_offsets:
                if fet_location < fet_offset:
                    # would lead to Blob underflow
                    continue
                if fet_location - fet_offset + rom_size > self.buffer_size:
                    # would lead to Blob overflow
                    continue
                try:
                    rom_offset = fet_location - fet_offset  # e.g. 0x20800 - 0x20000 = 0x0800
                    potential_rom = Rom(self, rom_size, rom_offset, fet_offset, psptool)
                    self.roms.append(potential_rom)
                    fet_parsed = True
                    break  # found correct fet_offset!
                except EmptyFet:
                    self.psptool.ph.print_warning(f"Empty FET at offset {hex(fet_offset)}, trying next offset")
                    continue
            if not fet_parsed:
                self.psptool.ph.print_warning(f"Skipping FET at {hex(fet_location)} due to unknown ROM alignment")

        if len(self.roms) == 0:
            self.psptool.ph.print_warning("Could not find any Firmware Entry Table!")

        self._construct_range_dict()

    def __repr__(self):
        return f'Blob({self.roms=})'

    def _construct_range_dict(self):
        all_entries = self.unique_entries()

        # create RangeDict in order to find entries, directories and fets for a given address
        directories = [directory for rom in self.roms for directory in rom.directories]
        self.range_dict = RangeDict({
            **{
                range(entry.get_address(),
                      entry.get_address() + entry.buffer_size):  # key is start and end address of the entry
                entry
                for entry in all_entries if entry.buffer_size != 0xffffffff  # value is its type
            }, **{
                range(directory.get_address(), directory.get_address() + len(directory)):
                    directory
                for directory in directories
            }, **{
                range(rom.fet.get_address(), rom.fet.get_address() + len(rom.fet)):
                    rom.fet
                for rom in self.roms
            }
        })

    def unique_entries(self) -> set:
        directories = [directory for rom in self.roms for directory in rom.directories]
        directory_entries = [directory.entries for directory in directories]
        # flatten list of lists
        all_entries = [entry for sublist in directory_entries for entry in sublist]
        # filter duplicates through set
        unique_entries = set(all_entries)
        return unique_entries

    def _find_fets(self):
        # AA55AA55 is to unspecific, so we require a word of padding before (to be tested)
        for m in re.finditer(b'\xff\xff\xff\xff' + self._FIRMWARE_ENTRY_MAGIC, self.get_buffer()):
            fet_offset = m.start() + 4
            yield fet_offset

    def _find_inline_pubkeys(self, fp):

        """ Try to find a pubkey anywhere in the blob.
        The pubkey is identified by its fingerprint. If found, the pubkey is
        added to the list of pubkeys of the blob """
        found_pubkeys = []

        m = re.finditer(re.escape(binascii.a2b_hex(fp)), self.get_bytes())
        for index in m:
            start = index.start() - 4
            if int.from_bytes(self[start:start + 4], 'little') == 1:
                # Maybe a pubkey. Determine its size:
                pub_exp_size = int.from_bytes(self[start + 0x38: start + 0x3c],
                                              'little')
                if pub_exp_size == 2048:
                    size = 0x240
                elif pub_exp_size == 4096:
                    size = 0x440
                else:
                    continue

                key_id = self[start + 0x04: start + 0x14]
                cert_id = self[start + 0x14: start + 0x24]

                if key_id != cert_id and cert_id != b'\0' * 0x10:
                    if pub_exp_size == 2048:
                        size += 0x100
                    else:
                        size += 0x200

                try:
                    entry = PubkeyEntry(self, self, 0xdead, size, start, self, self.psptool)
                    # todo: use from_fields factory instead of PubkeyEntry init
                    # entry = Entry.from_fields(self, self.parent_buffer,
                    #                           0xdead,
                    #                           size,
                    #                           start,
                    #                           self,
                    #                           self.psptool)
                    assert isinstance(entry, PubkeyEntry)
                    entry.is_inline = True
                    entry.parent_entry = self.range_dict[entry.get_address()]
                    if type(entry.parent_entry) == PubkeyEntry:
                        break
                    entry.parent_entry.inline_keys.add(entry)
                    found_pubkeys.append(entry)
                except Entry.ParseError as e:
                    self.psptool.ph.print_warning(f"_find_pubkey: Entry parse error at 0x{start:x}")
                    self.psptool.ph.print_warning(f'{e}')
                except Exception as e:
                    self.psptool.ph.print_warning(f"_find_pubkey: Error couldn't convert key at: 0x{start:x}")
                    self.psptool.ph.print_warning(f'{e}')

        return found_pubkeys

    def find_inline_pubkey_entries(self, ids):
        found_pkes = []
        for key_id in ids:
            found_pkes += self._find_inline_pubkeys(key_id)
        return found_pkes

    def get_entries_by_type(self, type_) -> List[Entry]:
        entries = []

        for rom in self.roms:
            for _dir in rom.directories:
                for entry in _dir.entries:
                    if entry.type == type_:
                        entries.append(entry)

        return entries



import re

from .fet import Fet, EmptyFet
from .utils import NestedBuffer, sole


class Rom(NestedBuffer):
    def __init__(self, parent_blob, rom_size, rom_offset, fet_offset, psptool):
        super().__init__(parent_blob, rom_size, rom_offset)

        # ROMs will be mapped into memory to fit the very end of the 32 bit memory
        #  -> most ROMs are 16 MB in size, so addresses are starting at 0xFF000000
        #  -> some ROMs are 8 MB in size, so addresses are starting at 0xFF800000
        self.addr_mask = rom_size - 1

        self.unique_entries = set()
        self.pubkeys = {}

        self.agesa_version = self._find_agesa_version()
        self.fet = Fet(self, fet_offset, psptool)

        self.directories = self.fet.directories

    def _find_agesa_version(self):
        # from https://www.amd.com/system/files/TechDocs/44065_Arch2008.pdf

        # todo: use NestedBuffers instead of saving by value

        m = re.compile(b"AGESA!..\x00.*?\x00")
        res = m.findall(self.get_bytes())

        if not res:
            return 'AGESA_UNKNOWN'
        else:
            res = sole(set(res), assert_msg=f"Found conflicting AGESA version strings: {res}!")

        return str(re.sub(b'\x00', b' ', res).strip().decode("ascii"))

        # # We are only interested in different agesa versions
        # res = set(res)
        # res = list(res)
        #
        # # Some Images contain actually two ROM images. I.e. one for Naples and
        # # one for Rome. Both will contain a valid FET which needs to be parsed.
        # if len(res) == 2:
        #     self.dual_rom = True
        #     self.agesa_version = str(re.sub(b'\x00', b' ', res[0]).strip().decode("ascii"))
        #     self.agesa_version_second = str(re.sub(b'\x00', b' ', res[1]).strip().decode("ascii"))
        # elif len(res) == 1:
        #     self.agesa_version = str(re.sub(b'\x00', b' ', res[0]).strip().decode("ascii"))
        # else:
        #     self.agesa_version = str("UNKNOWN")

    def __repr__(self):
        return f'Rom({self.agesa_version=})'


# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from binascii import hexlify
from os import listdir, mkdir, path

from abc import ABC, abstractmethod

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import load_pem_private_key, Encoding, PrivateFormat, BestAvailableEncryption, NoEncryption
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature

from .utils import NestedBuffer


# Abstract classes
class WithSignatureSize(ABC):

    # abstract static property
    def __init_subclass__(cls):
        cls.signature_size = cls._signature_size()

    @classmethod
    @abstractmethod
    def _signature_size(cls) -> int:
        pass


class PublicKey(WithSignatureSize):

    # abstract static property
    def __init_subclass__(cls):
        super().__init_subclass__()

    # to/from crypto material
    @classmethod
    @abstractmethod
    def from_crypto_material(cls, crypto_material: bytes):
        pass

    @abstractmethod
    def get_crypto_material(self, size: int) -> bytes:
        pass

    # core functionality
    @abstractmethod
    def verify_blob(self, blob: bytes, signature: bytes) -> bool:
        pass


class PrivateKey(WithSignatureSize):

    # abstract static property
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.PublicKey = cls._PublicKey()

    @staticmethod
    @abstractmethod
    def _PublicKey() -> PublicKey:
        pass

    # generate out of thin air
    @classmethod
    @abstractmethod
    def generate_new(cls):
        pass

    # to/from file
    @classmethod
    @abstractmethod
    def load_from_file(cls, filename: str, password: str = None):
        pass

    @abstractmethod
    def save_to_file(self, filename: str, password: str = None):
        pass

    # core functionality
    @abstractmethod
    def get_public_key(self) -> PublicKey:
        pass

    @abstractmethod
    def sign_blob(self, blob: bytes) -> bytes:
        pass



class KeyType:

    _key_types = dict()

    @staticmethod
    def from_name(name: str):
        return KeyType._key_types[name]

    def __init__(self, name: str, public_key_cls, private_key_cls):

        if KeyType._key_types.get(name):
            raise Exception(f'There is already a KeyType with the name "{name}"!')

        self.name = name

        assert issubclass(public_key_cls, PublicKey)
        assert issubclass(private_key_cls, PrivateKey)
        assert private_key_cls.signature_size == public_key_cls.signature_size

        self.signature_size = private_key_cls.signature_size
        self.PublicKey = public_key_cls
        self.PrivateKey = private_key_cls

        KeyType._key_types[name] = self


    def load_private_key(self, filename: str, password: str = None) -> PrivateKey:
        return self.PrivateKey.load_from_file(filename, password=password)

    def generate_private_key(self) -> PrivateKey:
        return self.PrivateKey.generate_new()

    def make_public_key(self, crypto_material: bytes) -> PublicKey:
        return self.PublicKey.from_crypto_material(crypto_material)


def _create_parent_dirname(dirname):
    dirname = path.dirname(dirname)
    if not path.exists(dirname):
        _create_parent_dirname(dirname)
        mkdir(dirname)

def create_parent_dir(filestub):
    _create_parent_dirname(path.realpath(filestub))


class PrivateKeyDict:

    def __init__(self, keys = dict()):
        self.keys = keys

    def __getitem__(self, name: str) -> PrivateKey:
        if not self.keys.get(name):
            self.keys[name] = KeyType.from_name(name).generate_private_key()
        return self.keys[name]

    def save_to_files(self, filestub: str, password: str = None):
        create_parent_dir(filestub)
        for (name, key) in self.keys.items():
            key.save_to_file(filestub + '.' + name, password=password)

    @staticmethod
    def read_from_files(filestub: str, password: str = None):
        create_parent_dir(filestub)
        dirname = path.dirname(filestub)
        filestub = path.relpath(filestub, start=dirname)
        keys = dict()
        for filename in listdir(dirname):
            if filename.startswith(filestub):
                name = filename.rsplit('.')[-1]
                keys[name] = KeyType.from_name(name).load_private_key(dirname + '/' + filename, password=password)
        return PrivateKeyDict(keys)


# Key Implementations


class RsaKey(WithSignatureSize):

    # abstract static property
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.key_bits = None
        if cls.signature_size:
            cls.key_bits = cls.signature_size << 3
            if cls.key_bits == 2048:
                cls.hash_algorithm = hashes.SHA256()
                cls.salt_length = 32
            elif cls.key_bits == 4096:
                cls.hash_algorithm = hashes.SHA384()
                cls.salt_length = 48
            else:
                raise Exception(f'Unknown rsa key size: {cls.key_bits} bits!')

    @classmethod
    def padding(cls):
        return padding.PSS(
            mgf=padding.MGF1(cls.hash_algorithm),
            salt_length=cls.salt_length
        )


class RsaPublicKey(PublicKey, RsaKey):

    def __init__(self, public_key):
        super().__init__()
        assert public_key.key_size == self.key_bits, f'Key has the wrong size: \
                expected {self.key_bits} but got {public_key.key_size}!'
        self._public_key = public_key

    # to/from crypto material
    @classmethod
    #override
    def from_crypto_material(cls, crypto_material: bytes) -> PublicKey:
        if len(crypto_material) == 2 * cls.signature_size:
            pubexp = int.from_bytes(crypto_material[:cls.signature_size], 'little')
            assert pubexp == 65537, f'The public exponent should always be 65537 \
                    (0x10001) not {pubexp} (0x{pubexp:x})!'
            modulus = int.from_bytes(crypto_material[cls.signature_size:], 'little')
        elif len(crypto_material) == cls.signature_size:
            pubexp = 0x10001
            modulus = int.from_bytes(crypto_material, 'little')
        else:
            raise Exception(f'Crypto material has unknown size: 0x{len(crypto_material):x}!')

        return cls(rsa.RSAPublicNumbers(pubexp, modulus).public_key())

    #override
    def get_crypto_material(self, size: int) -> bytes:
        numbers = self._public_key.public_numbers()
        modulus = numbers.n.to_bytes(self.signature_size, 'little')

        if size == self.signature_size:
            return modulus
        elif size == 2 * self.signature_size:
            pubexp = numbers.e.to_bytes(self.signature_size, 'little')
            return pubexp + modulus
        else:
            raise Exception(f'Unknown crypto_material size: 0x{size:x} \
                    (expected 0x{self.signature_size:x} or 0x{2*self.signature_size:x})!')

    # core functionality
    #override
    def verify_blob(self, blob: bytes, signature: bytes) -> bool:
        pass
        try:
            self._public_key.verify(
                signature,
                blob,
                self.padding(),
                self.hash_algorithm
            )
        except InvalidSignature:
            return False
        return True


class RsaPrivateKey(PrivateKey, RsaKey):

    def __init__(self, private_key):
        super().__init__()
        assert private_key.key_size == self.key_bits, f'Key has the wrong size: \
                expected {self.key_bits} but got {private_key.key_size}!'
        self._private_key = private_key

    # generate out of thin air
    @classmethod
    #override
    def generate_new(cls) -> PrivateKey:
        return cls(rsa.generate_private_key(public_exponent=0x10001,key_size=cls.key_bits))

    # to/from file
    @classmethod
    #override
    def load_from_file(cls, filename: str, password: str = None) -> PrivateKey:
        if password:
            password = password.encode()
        with open(filename, 'rb') as f:
            return cls(load_pem_private_key(f.read(), password=password))

    #override
    def save_to_file(self, filename: str, password: str = None):
        encryption = NoEncryption()
        if password:
            encryption=BestAvailableEncryption(password.encode())
        with open(filename, 'wb+') as f:
            f.write(self._private_key.private_bytes(
                encoding=Encoding.PEM,
                format=PrivateFormat.PKCS8,
                encryption_algorithm=encryption
            ))

    # core functionality
    #override
    def get_public_key(self) -> PublicKey:
        return self.PublicKey(self._private_key.public_key())

    #override
    def sign_blob(self, blob: bytes) -> bytes:
        return self._private_key.sign(
            blob,
            self.padding(),
            self.hash_algorithm
        )


class Rsa2048PublicKey(RsaPublicKey):

    def __init__(self, public_key):
        super().__init__(public_key)

    @classmethod
    #override
    def _signature_size(cls) -> int:
        return 0x100

class Rsa2048PrivateKey(RsaPrivateKey):

    def __init__(self, private_key):
        super().__init__(private_key)

    @staticmethod
    #override
    def _PublicKey() -> PublicKey:
        return Rsa2048PublicKey

    @classmethod
    #override
    def _signature_size(cls) -> int:
        return 0x100


rsa2048_key_type = KeyType("rsa2048", Rsa2048PublicKey, Rsa2048PrivateKey)


class Rsa4096PublicKey(RsaPublicKey):

    def __init__(self, public_key):
        super().__init__(public_key)

    @classmethod
    #override
    def _signature_size(cls) -> int:
        return 0x200

class Rsa4096PrivateKey(RsaPrivateKey):

    def __init__(self, private_key):
        super().__init__(private_key)

    @staticmethod
    #override
    def _PublicKey() -> PublicKey:
        return Rsa4096PublicKey

    @classmethod
    #override
    def _signature_size(cls) -> int:
        return 0x200

rsa4096_key_type = KeyType("rsa4096", Rsa4096PublicKey, Rsa4096PrivateKey)


class KeyId(NestedBuffer):

    @property
    def magic(self) -> str:
        return hexlify(self.get_bytes(0, 2)).upper().decode('ascii')

    def as_string(self) -> str:
        return hexlify(self.get_bytes()).upper().decode('ascii')

    def __repr__(self):
        return f'KeyId({self.as_string()})'


class Signature(NestedBuffer):
    @classmethod
    def from_nested_buffer(cls, nb):
        return Signature(nb.parent_buffer, nb.buffer_size, buffer_offset=nb.buffer_offset)


class ReversedSignature(Signature):
    def __getitem__(self, item):
        if isinstance(item, slice):
            new_slice = self._offset_slice(item)
            return self.parent_buffer[new_slice]
        else:
            assert (isinstance(item, int))
            assert item >= 0, "Negative index not supported for ReversedSignature"
            return self.parent_buffer[self.buffer_offset + self.buffer_size - item - 1]

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            new_slice = self._offset_slice(key)
            self.parent_buffer[new_slice] = value
        else:
            assert (isinstance(key, int))
            self.parent_buffer[self.buffer_offset + self.buffer_size - key - 1] = value

    def _offset_slice(self, item):
        return slice(
            self.buffer_offset + self.buffer_size - (item.start or 0) - 1,
            self.buffer_offset + self.buffer_size - (item.stop or self.buffer_size) - 1,
            -1
        )


# PSPTool - Display, extract and manipulate PSP firmware inside UEFI images
# Copyright (C) 2021 Christian Werling, Robert Buhren, Hans Niklas Jacob
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .entry import HeaderEntry, PubkeyEntry, KeyStoreEntry, KeyStoreKey
from .utils import NestedBuffer, RangeDict
from .crypto import PublicKey, PrivateKey, KeyType, PrivateKeyDict, KeyId, Signature

from . import errors

# Tree Types


class SignedEntity:
    def __init__(self, entry, certifying_id: KeyId, signature: Signature, psptool):
        self.psptool = psptool
        self.entry = entry
        self.certifying_id = certifying_id
        self.signature = signature

        # will be filled by CertificateTree
        self.certifying_keys = None
        self.contained_keys = set()

    def get_address(self):
        return self.entry.get_address()

    def get_length(self):
        return self.entry.buffer_size

    def get_range(self):
        start = self.get_address()
        return range(start, start + self.get_length())

    def __repr__(self) -> str:
        return f'SignedEntity(@{self.get_address():x}:{self.get_length():x})'

    @staticmethod
    def _from_pubkey_entry(pke, psptool):
        if pke.signed:
            return SignedEntity(pke, pke.certifying_id, pke.signature, psptool)
        return None

    @staticmethod
    def _from_header_entry(he: HeaderEntry, psptool):
        if not he.signed:
            return None

        signature = Signature.from_nested_buffer(he.signature)
        certifying_id = KeyId(he, 0x10, buffer_offset=0x38)

        return SignedEntity(he, certifying_id, signature, psptool)

    @classmethod
    def _from_key_store_entry(cls, kse: KeyStoreEntry, psptool):
        return SignedEntity(kse, kse.header.certifying_id, kse.signature, psptool)

    def is_verified(self):
        return self.verify_with_tree()

    def is_verified_by(self, pubkey):
        return self.verify_with_pubkey(pubkey.get_public_key())

    def verify_with_tree(self):
        if not self.certifying_keys:  # works for None or set()
            raise errors.NoCertifyingKey(self)

        verified_once = False
        failed_once = False
        for pubkey in self.certifying_keys:
            if self.verify_with_pubkey(pubkey.get_public_key()):
                verified_once = True
            else:
                failed_once = True

        if not verified_once:
            return False
        # todo: use this in strict mode (fails e.g. for H12SSW9.719)
        # if failed_once:
        #     raise errors.SignatureInconsistent(self, pubkey)
        return True

    def verify_with_pubkey(self, pubkey: PublicKey) -> bool:
        return pubkey.verify_blob(self.entry.get_signed_bytes(), self.signature.get_bytes())

    # resigns this entry only
    def resign_only(self, privkey: PrivateKey):
        print(f'Resigning {self} ({self.entry})')
        if self.entry.has_sha256_checksum:
            print(f'    Checking sha256 checksum of {self.entry}')
            if self.entry.verify_sha256(print_warning=False):
                print(f'        sha256 still valid!')
            else:
                print(f'        Need to rehash')
                self.entry.update_sha256()
                print(f'        Done')
        assert self.signature.buffer_size == privkey.signature_size, \
            f"{self.signature.buffer_size=}, {privkey.signature_size=}"
        signature = privkey.sign_blob(self.entry.get_signed_bytes())
        assert len(signature) == self.signature.buffer_size, f'Could not resign {self} with {privkey}: ' \
                                                             f'The new signature has the wrong length ' \
                                                             f'{len(signature)} != {self.signature.buffer_size}'
        self.signature.set_bytes(0, len(signature), signature)

    def resign_and_replace(self, privkeys: PrivateKeyDict = None, recursive: bool = False):
        # this resigns self (multiple times!)
        for pk in self.certifying_keys:
            if pk in self.contained_keys:
                continue # TODO hotfix
            pk.replace_and_resign(privkeys, recursive=recursive)


class PublicKeyEntity:
    def __init__(self, key_type: KeyType, key_id: KeyId, crypto_material: NestedBuffer, psptool):
        self.psptool = psptool
        self.key_type = key_type
        self.key_id = key_id
        self._crypto_material = crypto_material

        # will be filled by CertificateTree
        self.wrapping_entities = None
        self.certified_entities = None

        # for lazy loading
        self._public_key = None

    def get_address(self):
        return self.key_id.get_address()

    def get_magic(self):
        return self.key_id.as_string()[:4]

    def __repr__(self) -> str:
        return f'PubkeyEntity({self.get_magic()}, @{self.get_address():x})'

    def is_same(self, other) -> bool:
        if self.key_id[:] != other.key_id[:]:
            return False
        if self.key_type != other.key_type:
            return False
        if self._crypto_material[:] != other._crypto_material[:]:
            return False
        # TODO: we don't need this
        assert self.certified_entities == other.certified_entities
        return True

    @classmethod
    def _from_pubkey_entry(cls, pke: PubkeyEntry, psptool):
        if pke.modulus_size == 0x100:
            key_type = KeyType.from_name('rsa2048')
        elif pke.modulus_size == 0x200:
            key_type = KeyType.from_name('rsa4096')
        else:
            raise Exception(f'Unknown PubkeyEntry modulus size ({hex(pke.modulus_size)}) for {pke}')

        return PublicKeyEntity(key_type, pke.key_id, pke.crypto_material, psptool)

    @classmethod
    def _from_key_store_key(cls, ksk: KeyStoreKey, psptool):

        if ksk.key_size == 2048:
            key_type = KeyType.from_name('rsa2048')
        elif ksk.key_size == 4096:
            key_type = KeyType.from_name('rsa4096')
        else:
            raise Exception(f'Unknown key_size ({ksk.key_size:x}) for {ksk}')

        return PublicKeyEntity(key_type, ksk.key_id, ksk.crypto_material, psptool)

    def is_root(self) -> bool:
        return not self.get_certifying_keys() or self in self.get_certifying_keys()

    def get_certifying_ids(self):
        return set(entity.certifying_id
            for entity in self.wrapping_entities
        )

    def get_certifying_keys(self):
        return set(key
            for entity in self.wrapping_entities
                for key in entity.certifying_keys
        )

    def get_certified_keys(self):
        return set(key
            for entity in self.certified_entities
                for key in entity.contained_keys
        )

    def _make_public_key(self) -> PublicKey:
        try:
            return self.key_type.make_public_key(self._crypto_material.get_bytes())
        except:
            raise Exception(f'Cannot create crypto key for {self}.')

    def get_public_key(self) -> PublicKey:
        if not self._public_key:
            self._public_key = self._make_public_key()
        return self._public_key

    def replace_crypto_material(self, crypto_material: bytes):
        assert len(crypto_material) == self._crypto_material.buffer_size, f'Crypto material has wrong size: {len(crypto_material)} != {self._crypto_material.buffer_size}'
        self._public_key = None
        self._crypto_material.set_bytes(0, len(crypto_material), crypto_material)

    def replace_only(self, pubkey: PublicKey):
        print(f'Replacing {self}')
        assert self.key_type.signature_size == pubkey.signature_size
        size = self._crypto_material.buffer_size
        self.replace_crypto_material(pubkey.get_crypto_material(size))
        assert pubkey._public_key.public_numbers() == self.get_public_key()._public_key.public_numbers()

    def replace_and_resign(self, privkeys: PrivateKeyDict = None, recursive: bool = False):

        # get key
        if privkeys is None:
            privkeys = PrivateKeyDict()
        privkey = privkeys[self.key_type.name]
        assert self.key_type.signature_size == privkey.signature_size

        # replace self
        self.replace_only(privkey.get_public_key())

        # resign children
        for se in self.certified_entities:
            se.resign_only(privkey)

        # check crypto
        for se in self.certified_entities:
            assert se.is_verified_by(self), f'Resigning {se} with {self} failed!'

        # continue
        if recursive:
            if len(self.wrapping_entities) > 1:
                self.psptool.ph.print_warning(f'Resigning could be in wrong order for {self.wrapping_entities}!')

            for se in self.wrapping_entities:
                se.resign_and_replace(privkeys=privkeys, recursive=True)


class CertificateTree:
    def __init__(self, psptool):
        self.psptool = psptool
        # key_ids we have seen
        self.ids = set()
        # pubkeys by key id
        self.pubkeys = dict()
        # pubkeys by certifying_id address
        self.pubkeys_address = dict()
        # signed entities by certifying id
        self.signed_entities = dict()
        # signed entities by body address range
        self.signed_ranges = RangeDict()
        # signed entities where the verification failed
        self.verification_failed = set()

    def add_signed_entity(self, signed_entity: SignedEntity):
        address_range = signed_entity.get_range()

        # check if we already have this one
        if self.signed_ranges.get(address_range):
            raise errors.NonUniqueSignedEntity(self.signed_ranges[address_range], signed_entity)

        # add range->signed_entity dict
        self.signed_ranges[address_range] = signed_entity

        # update knowns key ids
        cert_id = signed_entity.certifying_id.as_string()
        self.ids.add(cert_id)

        # add key_id->signed_entity dict
        if not self.signed_entities.get(cert_id):
            self.signed_entities[cert_id] = set()
        self.signed_entities[cert_id].add(signed_entity)

        # update contained/wrapping edges
        for (address, pubkey) in self.pubkeys_address.items():
            if address in address_range:
                signed_entity.contained_keys.add(pubkey)
                pubkey.wrapping_entities.add(signed_entity)

        # update certifying edges
        certifying_keys = self.pubkeys.get(cert_id, set())
        assert signed_entity.certifying_keys is None
        signed_entity.certifying_keys = certifying_keys
        # update certified edges
        for pubkey in certifying_keys:
            pubkey.certified_entities.add(signed_entity)

    def add_pubkey_entity(self, pubkey: PublicKeyEntity):
        start_address = pubkey.get_address()

        # check if we already have this one
        if self.pubkeys_address.get(start_address):
            raise errors.NonUniquePublicKeyEntity(self.pubkeys_address[start_address], pubkey)

        # add address->pubkeys dict
        self.pubkeys_address[start_address] = pubkey

        # update known key ids
        key_id = pubkey.key_id.as_string()
        self.ids.add(key_id)

        # add key_id->pubkeys dict
        if not self.pubkeys.get(key_id):
            self.pubkeys[key_id] = set()
        self.pubkeys[key_id].add(pubkey)

        # update contained/wrapping edges
        assert pubkey.wrapping_entities is None
        pubkey.wrapping_entities = set()
        for (r, signed_entity) in self.signed_ranges.items():
            if start_address in r:
                pubkey.wrapping_entities.add(signed_entity)
                signed_entity.contained_keys.add(pubkey)

        # update certified edges
        assert pubkey.certified_entities is None
        pubkey.certified_entities = self.signed_entities.get(key_id, set())
        # update certifying edges
        for signed_entity in pubkey.certified_entities:
            signed_entity.certifying_keys.add(pubkey)

    def add_header_entry(self, header_entry) -> SignedEntity:
        signed_entity = SignedEntity._from_header_entry(header_entry, self.psptool)
        if signed_entity:
            try:
                self.add_signed_entity(signed_entity)
            except errors.NonUniqueSignedEntity as e:
                signed_entity = e.existing
            header_entry.signed_entity = signed_entity
            return signed_entity
        return None

    def add_pubkey_entry(self, pubkey_entry: PubkeyEntry) -> (PublicKeyEntity, SignedEntity):
        signed_entity = SignedEntity._from_pubkey_entry(pubkey_entry, self.psptool)
        if signed_entity:
            try:
                self.add_signed_entity(signed_entity)
            except errors.NonUniqueSignedEntity as e:
                signed_entity = e.existing
            pubkey_entry.signed_entity = signed_entity

        pubkey = PublicKeyEntity._from_pubkey_entry(pubkey_entry, self.psptool)
        if pubkey:
            try:
                self.add_pubkey_entity(pubkey)
            except errors.NonUniquePublicKeyEntity as e:
                pubkey = e.existing
            pubkey_entry.pubkey_entity = pubkey

        return pubkey, signed_entity

    def add_key_store_entry(self, key_store_entry: KeyStoreEntry):
        signed_entity = SignedEntity._from_key_store_entry(key_store_entry, self.psptool)
        if signed_entity:
            try:
                self.add_signed_entity(signed_entity)
            except errors.NonUniqueSignedEntity as e:
                signed_entity = e.existing
            key_store_entry.signed_entity = signed_entity

        for key_store_key in key_store_entry.key_store.keys:
            pubkey = PublicKeyEntity._from_key_store_key(key_store_key, self.psptool)
            try:
                self.add_pubkey_entity(pubkey)
            except errors.NonUniquePublicKeyEntity as e:
                pubkey = e.existing
            key_store_key.pubkey_entity = pubkey

    @staticmethod
    def from_blob(blob, psptool):
        ct = CertificateTree(psptool)

        for rom in blob.roms:
            for dr in rom.directories:
                for entry in dr.entries:
                    if type(entry) == PubkeyEntry:
                        ct.add_pubkey_entry(entry)
                    if type(entry) == HeaderEntry:
                        ct.add_header_entry(entry)
                    if type(entry) == KeyStoreEntry:
                        ct.add_key_store_entry(entry)

        # Add unlisted/inline keys as found by additional blob parsing efforts
        ids_to_find = ct.ids
        while ids_to_find:
            for pubkey in blob.find_inline_pubkey_entries(ids_to_find):
                ct.add_pubkey_entry(pubkey)
            ids_to_find = ct.ids.difference(ids_to_find)

        return ct

    def unique_pubkeys(self, key_id):
        unique_keys = list()
        for key in self.pubkeys[key_id]:
            inserted = False
            for uk in unique_keys:
                if key.is_same(uk[0]):
                    inserted = True
                    uk.append(key)
                    break
            if not inserted:
                unique_keys.append([key])
        return unique_keys

    def _print_key_tree_line(self, keys, indent):
        keys=list(keys)
        print(indent + f' +-{keys[0]}')
        keys=keys[1:]
        for key in keys:
            print(indent + f' | {key}')

    def _print_signed_entity_tree_line(self, signed_entity, verified, indent):
        print(indent + f' +-{signed_entity} (verified={verified})')

    def print_key_tree(self, root=None, indent='', stack=list()):
        seen_addresses = set()
        if not root:
            print(indent + 'AMD')
            for key_id in self.pubkeys.keys():
                for (keys) in self.unique_pubkeys(key_id):
                    key = keys[0]
                    if key.is_root() and key not in stack:
                        seen_addresses.update(set(map(lambda k: k.get_address(), keys)))
                        self._print_key_tree_line(keys, indent)
                        seen_addresses.update(self.print_key_tree(key, indent+' |', stack+[key]))
            #print("Seen:")
            #print(', '.join(map(hex,seen_addresses)))
            #print("Not seen:")
            #print(', '.join(map(hex,set(self.pubkeys_address) - seen_addresses)))
            assert seen_addresses == set(self.pubkeys_address.keys())
        else:
            for signed_entity in root.certified_entities:
                verified = signed_entity.is_verified_by(root)
                if not verified:
                    self.verification_failed.add(signed_entity)
                self._print_signed_entity_tree_line(signed_entity, verified, indent)
                for key in signed_entity.contained_keys:
                    if key in stack:
                        continue
                    key_id = key.key_id.as_string()
                    for keys in self.unique_pubkeys(key_id):
                        seen_addresses.update(set(map(lambda k: k.get_address(), keys)))
                        self._print_key_tree_line(keys, indent + ' |')
                    seen_addresses.update(self.print_key_tree(key, indent+' | |', stack+[key]))

        return seen_addresses
